"""Test pkcli to handle argh breaking changes

:copyright: Copyright (c) 2023 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def kwarg_to_positional(positional_arg, keyword_arg=None):
    return f'positional_arg={positional_arg} keyword_arg={keyword_arg}'



# -*- coding: utf-8 -*-
"""Create and read global and user manifests.

:copyright: Copyright (c) 2017 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

#: Appears in each directory
BASENAME = "rsmanifest.json"

# POSIT: https://github.com/radiasoft/containers/blob/master/bin/build build_rsmanifest()
#: Written once at build time
CONTAINER_FILE = "/" + BASENAME

#: Format version
FILE_VERSION = "20170217.180000"

#: Read and written multiple times as the run user
USER_FILE = "~/" + BASENAME

# Identifies codes which are not installed in a pyenv
_NO_PYENV = ""


def add_code(name, version, uri, source_d, virtual_env=None, pyenv=None):
    """Add a new code to ~?rsmanifest.json

    Args:
        name (str): name of the package
        version (str): commit or version
        uri (str): repo, source link
        source_d (str): directory containing
        virtual_env (str): DEPRECATED
        pyenv (str): pyenv version
    """
    from pykern import pkcollections
    from pykern import pkio
    from pykern import pkjson
    import datetime
    import json

    fn = pkio.py_path(USER_FILE)
    try:
        values = pkjson.load_any(fn)
    except Exception as e:
        if not (pkio.exception_is_not_found(e) or isinstance(e, ValueError)):
            raise
        values = pkcollections.Dict(
            version=FILE_VERSION,
            codes=pkcollections.Dict({_NO_PYENV: pkcollections.Dict()}),
        )
    if pyenv:
        assert not virtual_env, "only one of pyenv or virtual-env (DEPRECATED)"
    elif virtual_env:
        assert not pyenv, "only one of pyenv or virtual-env (DEPRECATED)"
        pyenv = virtual_env
    if not pyenv:
        pyenv = _NO_PYENV
    v = values.codes.get(pyenv) or pkcollections.Dict()
    v[name.lower()] = pkcollections.Dict(
        installed=datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
        name=name,
        source_d=source_d,
        uri=uri,
        version=version,
    )
    values.codes[pyenv] = v
    pkjson.dump_pretty(values, filename=fn)


def pkunit_setup():
    """Create rsmanifest files"""
    from pykern import pkjson

    pkjson.dump_pretty(
        {
            "version": FILE_VERSION,
            "image": {
                "type": "pkunit",
            },
        },
        filename=CONTAINER_FILE,
    )
    add_code("pkunit", "1.1", "https://pykern.org", "/tmp")


def read_all():
    """Merge all manifests

    Returns:
        dict: merged data
    """
    from pykern import pkio
    from pykern import pkjson

    fn = pkio.py_path(USER_FILE)
    # Both must exist or error
    u = pkjson.load_any(fn)
    c = pkjson.load_any(pkio.py_path(CONTAINER_FILE))
    assert u.version == c.version, "(user.version) {} != {} (container.version)".format(
        u.version, c.version
    )
    # There are "guaranteed" to be no collisions, but if there are
    # we override user.
    c.update(u)
    return c



"""Useful I/O operations

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

# Root module: Limit imports to avoid dependency issues
from pykern import pkconst
from pykern import pkinspect
import contextlib
import errno
import filecmp
import glob
import io
import os
import os.path
import py
import random
import re
import shutil

#: used during unit testing see ``pykern.pkunit.save_chdir``
pkunit_prefix = None

TEXT_ENCODING = "utf-8"


def atomic_write(path, contents, **kwargs):
    """Overwrites an existing file with contents via rename to ensure integrity

    Args:
        path (str or py.path.Local): Path of file to overwrite
        contents (str): New contents
        kwargs (kwargs): to pass to `py.path.local.write`
    """
    n = py_path(path).new(ext="pkio-tmp-" + random_base62())
    assert not n.exists(), f"{n} already exists (file name collision)"
    try:
        n.write(contents, **kwargs)
        n.rename(path)
    finally:
        # unchecked_remove is too brutal for this specific case
        if n.exists():
            try:
                os.remove(str(n))
            except Exception:
                pass


def compare_files(path1, path2, force=False):
    """Compares two files using `filecmp.cmp`

    Note that `filecmp` uses `os.stat` to see if a file is the
    same. If the size, mtime, and type are not identical, it does
    a comparison of the contents.

    `filecmp` caches prior resuls of the content comparisons.
    `force` "ensures" no caching, but since the cache is global,
    this can't be guaranteed in multithreaded environments.

    Args:
        path1 (str or py.path): first file
        path2 (str or py.path): second file
        force (bool): if True, call `filecmp.clear_cache` before comparison and ignore stats.

    Returns:
        bool: True if the files exist and have the same stats

    """
    if force:
        filecmp.clear_cache()
    try:
        return filecmp.cmp(str(path1), str(path2), shallow=not force)
    except Exception as e:
        if exception_is_not_found(e):
            return False
        raise


def exception_is_not_found(exc):
    """True if exception is one various file not found errors

    Checks `FileNotFoundError` and `IOError` with `errno.ENOENT`.

    Args:
        exc (BaseException): to check

    Returns:
        bool: True if is a file not found exception.
    """
    return (
        isinstance(exc, FileNotFoundError)
        or isinstance(exc, IOError)
        and exc.errno == errno.ENOENT
        or isinstance(exc, py.error.ENOENT)
    )


def has_file_extension(filename, to_check):
    """if matches any of the file extensions

    Args:
        filename (str|py.path.local): what to check
        to_check (str|tuple|list): is without '.' and lower

    Returns:
        bool: if any of the extensions matches
    """
    if isinstance(to_check, pkconst.STRING_TYPES):
        to_check = (to_check,)
    e = py_path(filename).ext[1:].lower()
    return e in to_check


def is_pure_text(filepath, test_size=512):
    """Read test_size bytes of filepath to determine if it is likely a text file.

    See `pykern.util.is_pure_text` for the heuristics used to test bytes.

    Args:
        filepath (str|py.path): file to check
        test_size (int): number of bytes to read from filename

    Returns:
        bool: True if file is likely pure text, false if likely binary
    """
    from pykern import util

    with open(filepath, "rb") as f:
        b = f.read(test_size + 1)
    return util.is_pure_text(b[:test_size], is_truncated=len(b) > test_size)


def mkdir_parent(path):
    """Create the directories and their parents (if necessary)

    Args:
        path (str): dir to create

    Returns:
        py.path.local: path
    """
    return py_path(path).ensure(dir=True)


def mkdir_parent_only(path):
    """Create the paths' parent directories.

    Args:
        path (str): children of dir to create

    Returns:
        py.path.local: parent directory of path
    """
    return mkdir_parent(py_path(path).dirname)


def open_text(filename, **kwargs):
    """Open file with utf-8 for text.

    Args:
        filename (str or py.path.Local): File to open

    Returns:
        object: open file handle
    """
    kwargs.setdefault("mode", "rt")
    kwargs.setdefault("encoding", TEXT_ENCODING)
    return io.open(str(py_path(filename)), **kwargs)


def py_path(path=None):
    """Creates a py.path.Local object

    Will expanduser, if needed.

    If `pkunit_prefix` is set, will prefix, too.

    Args:
        path (str): path to convert (or None for current dir)

    Returns:
        py.path.Local: path
    """
    global pkunit_prefix

    res = py.path.local(path, expanduser=True)
    if pkunit_prefix:
        # Allow for <test>_work and <test>_data so we don't add
        # prefix if there's a common parent directory.
        if not str(res).startswith(pkunit_prefix.dirname):
            res = pkunit_prefix.join(res)
        py.path.local(res.dirname).ensure(dir=True)
    return res


def random_base62(length=16):
    """Returns a safe string of sufficient length to be a nonce.

    The default is 62^16, which is 4e28. For comparison, 2^64 is 1e19 and
    2^128 is 3e38.

    Args:
        length (int): how long to make the base62 string [16]
    Returns:
        str: random base62 characters
    """
    r = random.SystemRandom()
    return "".join(r.choice(pkconst.BASE62_CHARS) for x in range(length))


def read_binary(filename):
    """Open file, read binary, and close.

    Args:
        filename (str or py.path.Local): File to open

    Returns:
        bytes: contents of `filename`
    """
    return py_path(filename).read_binary()


def read_text(filename):
    """Open file, read with utf-8 text, and close.

    Args:
        filename (str or py.path.Local): File to open

    Returns:
        Str: contents of `filename`
    """
    try:
        with open_text(filename) as f:
            return f.read()
    except Exception as e:
        pkinspect.append_exception_reason(e, f"filename={filename}")
        raise


@contextlib.contextmanager
def save_chdir(dirname, mkdir=False, is_pkunit_prefix=False):
    """Save current directory, change to directory, and restore.

    Args:
        dirname (str): directory to change to
        mkdir (bool): Make the directory?
        is_pkunit_prefix (bool): If True, sets pkunit_prefix.

    Returns:
        str: current directory before `chdir`
    """
    global pkunit_prefix

    prev_d = py.path.local().realpath()
    prev_ppp = pkunit_prefix
    try:
        if is_pkunit_prefix:
            d = py.path.local(dirname)
        else:
            d = py_path(dirname)
        if mkdir and not d.check(dir=1):
            mkdir_parent(d)
        os.chdir(str(d))
        if is_pkunit_prefix:
            pkunit_prefix = py.path.local(d)
        yield d.realpath()
    finally:
        os.chdir(str(prev_d))
        if is_pkunit_prefix:
            pkunit_prefix = prev_ppp


def sorted_glob(pattern, key=None):
    """Returns sorted list of files & dirs matching pattern.

    Use '**' in pattern for recursive search, else, use * as wildcard.
    Sorts using key if provided, else in ascending order.
    Doesn't include dot files unless dot "." is included explicitly
    at the start of a path component. Doesn't include . and ..
    To return files only, see walk_tree().

    Args:
        pattern (py.path.Local or str): to match file paths
        key (str): used to sort, must be name of py.path.Local attribute

    Returns:
        list: py.path.Local objects in sorted order
    """

    def _path_sort_attr(path):
        a = getattr(path, key)
        if callable(a):
            return a()
        return a

    return sorted(
        (py_path(f) for f in glob.iglob(str(pattern), recursive=True)),
        key=_path_sort_attr if key else None,
    )


def unchecked_remove(*paths):
    """Remove files or directories, ignoring OSError.

    Will not remove '/' or '.'

    Args:
        paths (str): paths to remove
    """
    cwd = py_path()
    for a in paths:
        p = py_path(a)
        assert len(p.parts()) > 1, "{}: will not remove root directory".format(p)
        assert cwd != p, "{}: will not remove current directory".format(p)
        try:
            os.remove(str(a))
        except OSError:
            try:
                shutil.rmtree(str(a), ignore_errors=True)
            except OSError:
                pass


def walk_tree(dirname, file_re=None):
    """Returns list of all files (only) in dirname (recursive), sorted in ascending order.

    Include file_re to filter results.
    Includes dot files, but not . and ..
    To include dirs in results, see sorted_glob().

    Args:
        dirname (str): top-level directory to walk
        file_re (re or str): Optionally, only return files which match the regular expression

    Returns:
        list: py.path.Local objects in sorted order
    """

    def _walk(dirname):
        for r, _, x in os.walk(dirname):
            r = py_path(r)
            for f in x:
                yield r.join(f)

    if not file_re:
        res = _walk(dirname)
    else:
        if not hasattr(file_re, "search"):
            file_re = re.compile(file_re)
        d = py_path(dirname)
        res = []
        for p in _walk(dirname):
            if file_re.search(d.bestrelpath(p)):
                res.append(p)
    return sorted(res)


def write_binary(path, contents):
    """Open file, write binary, and close.

    Args:
        path (str or py.path.Local): Path of file to write to
        contents (bytes): New contents

    Returns:
        py.path.local: `filename` as :class:`py.path.Local`
    """
    p = py_path(path)
    p.write_binary(contents)
    return p


def write_text(path, contents):
    """Open file, write text with utf-8, and close.

    Args:
        path (str or py.path.Local): Path of file to write to
        contents (str or bytes): New contents

    Returns:
        py.path.local: `filename` as :class:`py.path.Local`
    """
    from pykern import pkcompat

    p = py_path(path)
    try:
        with io.open(str(p), "wt", encoding=TEXT_ENCODING) as f:
            f.write(pkcompat.from_bytes(contents))
    except Exception as e:
        pkinspect.append_exception_reason(e, f"path={path}")
        raise
    return p



def args_and_imports(be, a, very, poorly, formatted, function, that, noone, should, ever, use, it, has, way, too, many, arguments):
    from pykern.pkdebug import pkdp
    x ='abcdefg'



    def foo():
        return 1
    def bar(): return 3



# -*- coding: utf-8 -*-
u"""test `pykern.pkconfig`

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
from pykern import pkconfig
from pykern.pkdebug import pkdc, pkdp
import os

def _custom_p6(v):
    import dateutil.parser
    return dateutil.parser.parse(v)

@pkconfig.parse_none
def _some_key(v):
    if v is None:
        return 999
    return int(v)

cfg = pkconfig.init(
    dict1=({
        'd1': 'default1',
        'd2': 'default2',
    }, dict, 'first param is dict'),
    list2=(['second1'], list, 'second param is list'),
    p3=(1313, int, 'third param is int'),
    p4=(None, int, 'fourth param is 10x p3'),
    p6=(None, _custom_p6, 'sixth param is a custom parser'),
    list7=(['default7'], list, 'seventh param is a list '),
    req8=pkconfig.Required(int, 'an eighth required parameter'),
    sub_params9=dict(
        sub9_1=(None, int, 'sub param is first of ninth group'),
        sub9_2=dict(
            sub9_2_1=(44, int, 'sub 9.2.1')
        ),
    ),
    req10=pkconfig.RequiredUnlessDev(('dev-host',), tuple, 'a dev default-only param'),
    dynamic_default10=(None, _some_key, 'sub dynamic default by parsing None'),
    bool1=(False, bool, 'a False boolean'),
    bool2=(True, bool, 'a True boolean'),
    bool3=(True, bool, 'a True boolean will be overriden'),
    bool4=(False, bool, 'a False boolean will be overriden'),
    tuple1=((), tuple, 'an empty tuple'),
    tuple2=((1,), tuple, 'a single value tuple'),
    tuple3=((2,), tuple, '(2,) will be overrriden'),
    tuple4=((3,), tuple, '(3,) will be overriden'),
    set1=(set(), set, 'an empty set'),
    # sets should have only one value of a type
    set2=((1, 1), set, 'a single value set'),
    set3=(set([2,]), frozenset, '(2,) will be overrriden'),
    set4=((3,), set, '(3,) will be overriden'),
)
if cfg.p4 is None:
    cfg.p4 = cfg.p3 * 10



"""Invoke commands from command line interpreter modules.

Any module in ``<root_pkg>.pkcli`` will be found by this module. The
public functions of the module will be executed when called from the
command line. This module is invoked by :mod:`pykern.pykern_console`.
Every project must have its own invocation module.

The basic form is: <project> <simple-module> <function>. <simple-module>
is the module without `<root_pkg>.pkcli`.  <function> is any function
that begins with a letter and contains word characters (\w).

If the module only has one public function named default_command,
the form is: <project> <simple-module>.

The purpose of this module is to simplify command-line modules. There is
no boilerplate. You just create a module with public functions
in a particular package location (e.g. `pykern.pkcli`).
This module does the rest.

`pykern.pkcli.pkexample` is a working example.

:copyright: Copyright (c) 2015-2023 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

import argh
import argh.assembling
import argparse
import importlib
import inspect
import os
import os.path
import pkgutil
import re
import sys

# Avoid pykern imports so avoid dependency issues for pkconfig
from pykern import pkconfig
from pykern import pkconst
from pykern import pkinspect

#: Sub-package to find command line interpreter (cli) modules will be found
CLI_PKG = ["pkcli"]

#: If a module only has one command named this, then execute directly.
DEFAULT_COMMAND = "default_command"

#: Test for first arg to see if user wants help
_HELP_RE = re.compile(r"^-(-?help|h)$", flags=re.IGNORECASE)

#: Used by `_fix_sys_path`
_fix_sys_path_done = False


class CommandError(Exception):
    """argh.CommandError is caught by argh, and does not cause an exit

    This CommandError causes an exit(1).
    """

    pass


def command_error(fmt, *args, **kwargs):
    """Raise CommandError with msg

    Args:
        fmt (str): how to represent arguments

    Raises:
        CommandError: always
    """
    raise CommandError(fmt.format(*args, **kwargs))


def command_info(fmt, *args, **kwargs):
    """Write message to stderr without raising

    Args:
        fmt (str): how to represent arguments
    """
    sys.stderr.write(fmt.format(*args, **kwargs))


class CustomFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter
):
    def _expand_help(self, action):
        return super()._expand_help(action).split("\n")[0]


class CustomParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.program = kwargs.copy()
        self.options = []

    def format_help(self):
        f = argh.PARSER_FORMATTER(prog=self.prog)
        if not self.description:
            f = CustomFormatter(prog=self.prog)
        f.add_usage(self.usage, self._actions, self._mutually_exclusive_groups)
        f.add_text(self.description)
        for a in self._action_groups:
            f.start_section(a.title)
            f.add_text(a.description)
            f.add_arguments(a._group_actions)
            f.end_section()
        f.add_text(self.epilog)
        if not self.description:
            return f.format_help().replace("positional arguments", "commands")
        return f.format_help()

    def print_help(self):
        pkconst.builtin_print(self.format_help())


def main(root_pkg, argv=None):
    """Invokes module functions in :mod:`pykern.pkcli`

    Looks in ``<root_pkg>.pkcli`` for the ``argv[1]`` module. It then
    invokes the ``argv[2]`` method of that module.

    Args:
        root_pkg (str): top level package name
        argv (list of str): Defaults to `sys.argv`. Only used for testing.

    Returns:
        int: 0 if ok. 1 if error (missing command, etc.)
    """
    _fix_sys_path()
    pkconfig.append_load_path(root_pkg)
    if not argv:
        argv = list(sys.argv)
    prog = os.path.basename(argv.pop(0))
    if _is_help(argv):
        return _list_all(root_pkg, prog)
    module_name = argv.pop(0)
    cli = _module(root_pkg, module_name)
    if not cli:
        return 1
    prog = prog + " " + module_name
    parser = CustomParser(prog)
    cmds = _commands(cli)
    dc = _default_command(cmds, argv)
    if dc:
        argh.set_default_command(
            parser,
            dc,
            **_argh_name_mapping_policy(),
        )
    else:
        argh.add_commands(
            parser,
            cmds,
            **_argh_name_mapping_policy(),
        )
        if len(argv) < 1:
            # Python 3: parser doesn't exit if not enough commands
            parser.error("too few arguments")
        if argv[0][0] != "-":
            argv[0] = _module_to_cmd(argv[0])
    from pykern.pkdebug import pkdp

    try:
        res = argh.dispatch(parser, argv=argv)
    except CommandError as e:
        sys.stderr.write("error: {}\n".format(e))
        return 1
    return 0


def _argh_name_mapping_policy():
    """Return name_mapping_policy based on feature test of NameMappingPolicy.

    In v0.30.0 argh changed the policy for mapping function args to
    CLI args. For versions 0.30.0 and above this sets the policy to
    BY_NAME_IF_HAS_DEFAULT which most closely matches the behavior
    prior to v0.30.0.
    """
    p = getattr(argh.assembling, "NameMappingPolicy", None)
    if not p:
        # Prior to v0.30.0 there was no NameMappingPolicy policy so no
        # need to set.
        return {}
    return {"name_mapping_policy": p.BY_NAME_IF_HAS_DEFAULT}


def _commands(cli):
    """Extracts all public functions from `cli`

    Args:
        cli (module): where commands are executed from

    Returns:
        list of function: public functions sorted alphabetically
    """
    res = []
    for n, t in inspect.getmembers(cli):
        if _is_command(t, cli):
            res.append(t)
    sorted(res, key=lambda f: f.__name__.lower())
    return res


def _default_command(cmds, argv):
    """Evaluate the default command, handling ``**kwargs`` case.

    `argparse` and `argh` do not understand ``**kwargs``, i.e. pass through command.
    There's a case (`pykern.pkcli.pytest`) that requires pass through so we wrap
    the command and clear `argv` in the case of ``default_command(*args, **kwargs)``.

    Args:
        cmds (list): List of commands
        argv (list): arguments (may be edited)

    Returns:
        function: default command or None
    """
    if len(cmds) != 1 or cmds[0].__name__ != DEFAULT_COMMAND:
        return None
    dc = cmds[0]
    spec = inspect.getargspec(dc)
    if not (spec.varargs and spec.keywords):
        return dc
    save_argv = argv[:]

    def _wrap_default_command():
        return dc(*save_argv)

    del argv[:]
    return _wrap_default_command


def _import(root_pkg, name=None):
    """Dynamically imports ``root_pkg.CLI_PKG[.name]``.

    Args:
        root_pkg (str): top level package
        name (str): cli module

    Returns:
        module: imported module

    Raises:
        ImportError: if module could not be loaded
    """

    def _imp(path_list):
        return importlib.import_module(_module_name(path_list))

    # TODO(robnagler) remove once all clients support pkcli directory
    path = None
    first_e = None
    m = None
    for p in CLI_PKG:
        path = [root_pkg, p]
        try:
            m = _imp(path)
            break
        except ImportError as e:
            # Assumes package (foo.pkcli) has an empty __init__.py so that
            # the import should always succeed.
            if not first_e:
                first_e = e
    if not path:
        raise first_e
    if not name:
        return m
    return _imp(path + [name])


def _is_command(obj, cli):
    """Is this a valid command function?

    Args:
        obj (object): candidate
        cli (module): module to which function should belong

    Returns:
        bool: True if obj is a valid command
    """
    if not inspect.isfunction(obj) or obj.__name__.startswith("_"):
        return False
    return hasattr(obj, "__module__") and obj.__module__ == cli.__name__


def _is_help(argv):
    """Does the user want help?

    Args:
        argv (list): list of args

    Returns:
        bool: True if no args or --help
    """
    if len(argv) == 0:
        return True
    return _HELP_RE.search(argv[0])


def _list_all(root_pkg, prog):
    """Prints a list of importable modules and exits.

    Searches ``<root_pkg>.pkcli` for submodules, and prints their names.

    Args:
        root_pkg (str): top level package
        prog (str): argv[0], name of program invoked

    Returns:
        int: 0 if ok. 1 if error.

    """
    res = "\n".join(
        sorted(
            pkinspect.package_module_names(_import(root_pkg)),
            key=str.lower,
        ),
    )
    sys.stderr.write(f"usage: {prog} module command [args...]\nModules:\n{res}\n")
    return 1


def _module(root_pkg, name):
    """Imports the module, catching `ImportError`

    Args:
        root_pkg (str): top level package
        name(str): unqualified name of the module to be imported

    Returns:
        module: imported module
    """

    def _match_exc(e):
        return re.search(
            " {}$|{}".format(
                # py2
                _module_from_cmd(name),
                # py3
                _module_name((root_pkg, name)),
            ),
            str(e),
        )

    try:
        return _import(root_pkg, name)
    except Exception as e:
        if (
            isinstance(e, ImportError)
            and _match_exc(e)
            or isinstance(e, (argh.CommandError, CommandError))
        ):
            sys.stderr.write(str(e) + "\n")
        else:
            raise
    return None


def _module_from_cmd(cmd):
    return cmd.replace("-", "_")


def _module_name(path_list):
    return _module_from_cmd(".".join(path_list))


def _module_to_cmd(module):
    return module.replace("_", "-")


def _fix_sys_path():
    """Remove the script directory from `sys.path`

    The script's directory is added to sys.path by the interpreter. In
    Python 3.11, this will change with the ``-P`` flag.
    """
    global _fix_sys_path_done

    if _fix_sys_path_done:
        return
    _fix_sys_path_done = True
    if not (sys.argv and sys.path):
        # Not enough information
        return
    d = os.path.dirname(os.path.realpath(sys.argv[0]))
    if sys.path[0] == d:
        sys.path.pop(0)



"""test pkcli.sphinx

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

_PROJECT = "xyzzy"
_VERSION = "20200202.123456"
_AUTHOR = "Arthur Coder"


def test_build(monkeypatch):
    from pykern import pkunit, pkio

    _mock_metadata(monkeypatch)
    with pkio.save_chdir(pkunit.empty_work_dir().join(_PROJECT).ensure(dir=True)):
        _projex()
        from pykern.pkcli import sphinx

        sphinx.build()
        pkunit.pkre(
            f"{_PROJECT} {_VERSION}", pkio.read_text("_build_html/html/index.html")
        )


def _mock_metadata(monkeypatch):
    from importlib import metadata

    def _mock(*args, **kwargs):
        nonlocal _prev
        if not args or args[0] != _PROJECT:
            return _prev(*args, **kwargs)
        # Do not use PKDict because that's not what metadata returns
        return {
            "Author-email": f"{_AUTHOR} <a@b.c>",
            "Name": _PROJECT,
            "Version": _VERSION,
        }

    _prev = metadata.metadata
    monkeypatch.setattr(metadata, "metadata", _mock)


def _projex():
    from pykern.pkcli import projex
    import subprocess

    subprocess.check_call(("git", "init", "."), stdout=subprocess.DEVNULL)
    projex.init_rs_tree(f"{_PROJECT} for sphinx_test")



# -*- coding: utf-8 -*-
"""Front-end command line for :mod:`pykern.pkcli`.

Example:

:copyright: Copyright (c) 2015 Bivio Software, Inc.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function

import sys

from pykern import pkcli


def main():
    return pkcli.main("pykern")


if __name__ == "__main__":
    sys.exit(main())



import pytest


@pytest.fixture(scope="function")
def pkconfig_setup(monkeypatch):
    import py.path
    import sys

    def res(cfg=None, env=None):
        # Can't import anything yet
        data_dir = py.path.local(__file__).dirpath("pkconfig_data")
        for k, v in (env or {}).items():
            monkeypatch.setenv(k, v)
        if data_dir not in sys.path:
            sys.path.insert(0, str(data_dir))
        from pykern import pkconfig

        pkconfig.reset_state_for_testing(add_to_environ=cfg)
        return pkconfig

    return res



class C():
    pass

v = 1
c = C()

from pykern import pkinspect


def caller(ignore_modules=None):
    return pkinspect.caller(ignore_modules=ignore_modules)


def caller_module(exclude_first=True):
    return pkinspect.caller_module(exclude_first=exclude_first)


def is_caller_main():
    return pkinspect.is_caller_main()



# -*- coding: utf-8 -*-
"""Various constant values

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

# pykern uses pksetup in setup.py so requirements.txt is not yet evaluated so can't use six
# or any other external dependencies. The "types" modules in Python 2 had a StringTypes, which
# would have been great if it hadn't been removed and changed completely in Python 3.

#: Python version independent value of string instance check
STRING_TYPES = None
try:
    STRING_TYPES = basestring
except NameError:
    STRING_TYPES = str


try:
    # May not be here, but that's ok, only for modules that
    # aren't in the critical import path (see comment above)
    import py.path

    #: Class of `pkio.py_path` and `py.path.local`
    PY_PATH_LOCAL_TYPE = type(py.path.local())
except Exception:
    pass


#: Copied from numconv, which copied from RFC1924
BASE62_CHARS = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"

#: The subdirectory in the top-level Python where to put resources
PACKAGE_DATA = "package_data"

#: Use this (sparingly, pkdlog is prefered) when you want to use print directly.
builtin_print = print



# -*- coding: utf-8 -*-
"""test pykern.pkcli.rsmanifest

:copyright: Copyright (c) 2017 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest


def test_add_code():
    from pykern import pkio
    from pykern import pkjson
    from pykern import pkunit
    from pykern.pkunit import pkok, pkeq, pkre
    from pykern.pkdebug import pkdp
    from pykern.pkcli import rsmanifest
    import re

    with pkunit.save_chdir_work(is_pkunit_prefix=True) as d:
        rsmanifest.add_code("A", "b", "c", "d", pyenv="v")
        j = pkjson.load_any(pkio.py_path(rsmanifest.USER_FILE).read())
        pkok(20170101.0 < float(j.version), "version must be after 2017")
        pkeq("A", j.codes.v.a.name)
        pkeq("b", j.codes.v.a.version)
        rsmanifest.add_code("a", "bb", "cc", "dd")
        j = pkjson.load_any(pkio.py_path(rsmanifest.USER_FILE).read())
        pkeq("A", j.codes.v.a.name)
        pkeq("a", j.codes[""].a.name)
        pkeq("bb", j.codes[""].a.version)
        pkre("20.*T.*Z", j.codes[""].a.installed)


def test_read_all():
    from pykern import pkio
    from pykern import pkjson
    from pykern import pkunit
    from pykern.pkunit import pkok, pkeq, pkre
    from pykern.pkdebug import pkdp
    from pykern.pkcli import rsmanifest
    import re

    with pkunit.save_chdir_work(is_pkunit_prefix=True) as d:
        rsmanifest.add_code(
            "code1",
            version="1.1",
            uri="http://x.com",
            source_d="/tmp",
            pyenv="py2",
        )
        v = pkjson.load_any(pkio.py_path(rsmanifest.USER_FILE)).version
        pkjson.dump_pretty(
            {"version": v, "image": {"type": "docker"}},
            filename=rsmanifest.CONTAINER_FILE,
        )
        m = rsmanifest.read_all()
        pkeq(v, m.version)
        pkeq("docker", m.image.type)
        pkeq("1.1", m.codes.py2.code1.version)



# -*- coding: utf-8 -*-
u"""Front-end command line for :mod:`pksetupunit2`.

See :mod:`pykern.pkcli` for how this module is used.

:copyright: Copyright (c) 2016 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function

import sys

from pykern import pkcli


def main():
    return pkcli.main('pksetupunit2')


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
import pykern.pksetup

pykern.pksetup.setup(
    name='pksetupunit1',
    description='pksetup Conformance One',
    author='RadiaSoft LLC.',
    author_email='pip@pykern.org',
    url='http://pykern.org',
    pksetup={
        'extra_directories':['examples'],
    },
)



# -*- coding: utf-8 -*-
"""pkconfig to_environ() test

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest


def test_to_environ(pkconfig_setup):
    pkconfig = pkconfig_setup(
        cfg=dict(OTHER_THING="", P1_M1_SET4="a:b"),
        env=dict(P1_M1_REQ8="33", PYKERN_PKCONFIG_DEV_MODE="0"),
    )

    from pykern.pkcollections import PKDict

    assert PKDict(P1_M1_REQ8="33", P1_M1_SET4="a:b") == pkconfig.to_environ(["p1.*"])
    assert PKDict(
        P1_M1_REQ8="33", P1_M1_SET4="a:b", OTHER_THING=""
    ) == pkconfig.to_environ(["p1.*", "other.thing"])
    assert PKDict() == pkconfig.to_environ(["nomatch.*"])

    from pykern import pkunit

    a = pkconfig.to_environ(
        ["foo.*", "baz.*.*"],
        values=dict(
            {
                "foo_bar": ["a", "c"],
                "foo.bar2": "2",
                "foo_BAR3": True,
                "FOO": {
                    "BAR4": False,
                    "BAR5": 5,
                },
                "baz.bar.exclude": 7,
                "baz.nomatch": 8,
                "baz.bar.foo": 9,
                "nomatch.foo": "4",
            }
        ),
        exclude_re="exclude",
    )
    pkunit.pkeq(
        dict(
            FOO_BAR="a:c",
            FOO_BAR2="2",
            FOO_BAR3="1",
            FOO_BAR4="",
            FOO_BAR5="5",
            BAZ_BAR_FOO="9",
        ),
        a,
    )
    pkunit.pkeq(False, pkconfig.in_dev_mode())



# -*- coding: utf-8 -*-
"""pkconfig init

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest


def test_init(pkconfig_setup):
    """Validate parse_set"""
    pkconfig = pkconfig_setup(
        cfg=dict(
            PYKERN_PKCONFIG_CHANNEL="dev",
            PYKERN_PKCONFIG_DEV_MODE="false",
            P1_M1_SET3="",
            P1_M1_SET4="a:b",
        ),
        env=dict(P1_M1_REQ8="99", P1_M1_HOSTS="h1"),
    )
    from pykern import pkunit

    with pkunit.pkexcept("p1.m1.req10: config"):
        from p1.m1 import cfg



# -*- coding: utf-8 -*-
"""Excel spreadsheet generator

:copyright: Copyright (c) 2021 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdc, pkdlog, pkdp
import decimal
import xlsxwriter
import re


_XL_COLS = None

_NO_CHILDREN = tuple()

_DEFAULT_ROUND_DIGITS = 2

_DIGITS_TO_PLACES = None

_SPACES = re.compile(r"\s+")

#: For sanity sake, we start at 1 with column numbers
_COL_NUM_1 = 1

_ROW_NUM_1 = 1

#: max rows is 1048576 so just use 10M. Used for sort_index
_ROW_MODULUS = 10000000

_OP_MULTI = PKDict(
    {
        "*": PKDict(xl="PRODUCT", func=lambda x, y: x * y, init=1),
        "+": PKDict(xl="SUM", func=lambda x, y: x + y, init=0),
    }
)
_OP_BINARY = PKDict(
    {
        "-": lambda x, y: x - y,
        "/": lambda x, y: x / y,
    }
)
_OP_UNARY = PKDict(
    {
        "-": lambda x: -x,
    }
)


class _Base(PKDict):
    def __init__(self, cfg):
        self.pkupdate(cfg).pksetdefault(defaults=PKDict)

    def cell(self, content_or_cell, **kwargs):
        """Convert content or a cell config to a `_Cell`

        If `content_or_cell` is a PKDict, is used to configure a cell. `kwargs` must be empty.

        If `content_or_cell` is a `_Cell`, returns itself. `kwargs` must be empty.

        Otherwise, `content_or_cell` is treated as content and `_Cell` is created with the content and `kwargs`.

        Args:
            content_or_cell (object): see above

        Returns:
            _Cell: instance
        """
        if isinstance(content_or_cell, _Cell):
            assert not kwargs
            return content_or_cell
        elif isinstance(content_or_cell, dict):
            assert not kwargs
            return _Cell(content_or_cell)
        else:
            kwargs["content"] = content_or_cell
            return _Cell(kwargs)

    def _cascade_defaults(self, parent_defaults):
        self.defaults.pksetdefault(**parent_defaults)
        for c in self._children():
            c._cascade_defaults(self.defaults)
        # TODO(robnagler): document clearing defaults with None
        for k in [k for k, v in self.defaults.items() if v is None]:
            del self.defaults[k]

    def _child(self, children, child, kwargs):
        s = child(kwargs)
        s._relations(self)
        children.append(s)
        return s

    def _compile_pass1(self):
        for c in self._children():
            c._compile_pass1()

    def _compile_pass2(self):
        for c in self._children():
            c._compile_pass2()

    def _error(self, fmt, *args, **kwargs):
        kwargs["self"] = self
        pkdlog(fmt + "; {self}", *args, **kwargs)
        # TODO: print stack
        raise AssertionError(f"workbook save failed; path={self.workbook.path}")

    def _relations(self, parent):
        self.parent = parent
        if isinstance(parent, _Sheet):
            self.sheet = parent
            self.workbook = parent.workbook
        elif isinstance(parent, Workbook):
            self.workbook = parent
        else:
            self.sheet = parent.sheet
            self.workbook = parent.workbook
        return self

    def pkdebug_str(self):
        l = ""
        for x in "link", "value", "xl_id":
            l += f",{x}={self[x]}" if x in self else ""
        for x in "content", "title", "path":
            if x in self:
                return f"{self.__class__.__name__}({x}={self[x]}{l})"
        return f"{self.__class__.__name__}({l})"

    def _print(self):
        pkdlog(self)
        for c in self._children():
            c._print()


class Workbook(_Base):
    def __init__(self, **kwargs):
        """Creates Workbook

        Args:
            path (py.path): where to write the spreadsheet
            defaults (PKDict): default values, e.g. round_digits
        """
        super().__init__(kwargs)
        self.sheets = []
        # for consistency in error messages
        self.workbook = self
        self.links = PKDict()

    def xl_fmt(self, cfg):
        """Get the Excel format for cfg

        Args:
            cfg (PKDict): key values that are supported by xlsxwriter

        Returns:
            Format: object which represents format
        """
        k = str(cfg)
        return self._xl_fmt.pksetdefault(k, lambda: self.xl.add_format(cfg))[k]

    def sheet(self, **kwargs):
        """Append a sheet to a Workbook

        Args:
            title (str): label for the sheet
            defaults (PKDict): default values, e.g. round_digits
        """
        return self._child(self.sheets, _Sheet, kwargs)

    def save(self):
        self._cascade_defaults(PKDict(str_fmt="text", num_fmt="decimal"))
        self._compile_pass1()
        self._compile_pass2()
        self._xl_fmt = PKDict()
        self.xl = xlsxwriter.Workbook(str(self.path))
        for s in self.sheets:
            s._save()
        self.xl.close()
        self._xl_fmt = None
        self.xl = None

    def _assert_link_pair(self, link, left, right):
        e = None
        for x in (
            ["fmt", left.get("fmt"), right.get("fmt")],
            ["is_decimal", left.is_decimal, right.is_decimal],
            ["type", type(left.value), type(right.value)],
            ["round_digits", left.round_digits, right.round_digits],
        ):
            if x[1] != x[2]:
                self._error(
                    "link={} {} {}={} different from {} {}={}",
                    link,
                    left,
                    x[0],
                    x[1],
                    right,
                    x[0],
                    x[2],
                )

    def _children(self):
        return self.sheets

    def _compile_pass2(self):
        super()._compile_pass2()
        for k, v in self.links.items():
            for i in range(len(v) - 1):
                self._assert_link_pair(k, v[i], v[i + 1])


class _Cell(_Base):
    def __init__(self, kwargs):
        super().__init__(kwargs)
        self._is_expr = False

    def _children(self):
        return _NO_CHILDREN

    def _compile_pass1(self):
        self._compile_link1()

    def _compile_pass2(self):
        if "is_compiled" in self:
            if not self.is_compiled:
                self._error("circular referenced cell={}", self)
            return
        self.is_compiled = False
        self._compile_content()
        self.is_compiled = True

    def _compile_content(self):
        if self.content is None:
            self._compile_str("")
            return
        elif isinstance(self.content, (float, int)):
            self.content = decimal.Decimal(self.content)
        if isinstance(self.content, str):
            self._compile_str(self.content)
            self.content = self.value
        elif isinstance(self.content, decimal.Decimal):
            self._compile_decimal(self.content)
            self.content = self.value
        elif isinstance(self.content, (list, tuple)):
            self._compile_expr_root()
        else:
            self._error(
                "content type={} not supported",
                type(self.content),
                self,
            )

    def _compile_decimal(self, value):
        # TODO(robnagler)
        # maybe this should not be applied because references should get to override
        # or maybe this should be overridden in the reference compilation always?
        # or maybe it takes the min of round_digits
        self.pksetdefault(
            round_digits=lambda: self.defaults.get(
                "round_digits", _DEFAULT_ROUND_DIGITS
            ),
        )
        self.value = _rnd(value, self.round_digits)
        self.content = str(self.value)
        self.is_decimal = True

    def _compile_expr(self, expr):
        if len(expr) == 0:
            self._error("empty expr")
        e = expr[0]
        if isinstance(e, (float, int, decimal.Decimal)):
            v = decimal.Decimal(e)
            return _Operand(
                content=str(v),
                count=1,
                fmt=None,
                round_digits=None,
                value=v,
                is_decimal=True,
            )
        # TODO(robnagler) document that links must begin with alnum
        if e[0][0].isalnum():
            return self._compile_ref(e, expect_count=None)
        return self._compile_op(expr)

    def _compile_expr_root(self):
        self._is_expr = True
        r = self._compile_expr(self.content)
        # expression's value overrides defaults
        for x in "fmt", "round_digits":
            if r.get(x) is not None:
                self.pksetdefault(x, r[x])
        if isinstance(r.value, decimal.Decimal):
            self._compile_decimal(r.value)
            self.content = f"=ROUND({r.content}, {self.round_digits})"
        elif isinstance(r.value, str):
            self._compile_str(r.value)
            self.content = f"={r.content}"
        else:
            raise AssertionError(f"_compile_expr invalid r={r}")

    def _compile_link1(self):
        if "link" in self:
            for l in [self.link] if isinstance(self.link, str) else self.link:
                if not l[0].isalnum():
                    self._error("link={} must begin with alphanumeric", l)
                self.workbook.links.setdefault(l, []).append(self)

    def _compile_op(self, expr):
        o = expr[0]
        e = expr[1:]
        if o == "+":
            return self._compile_operands(o, e, expect_count=None)
        if o == "-":
            if len(expr) > 2:
                return self._compile_operands(o, e, expect_count=2)
            return self._compile_operands(o, e, expect_count=1)
        if o == "*":
            return self._compile_operands(o, e, expect_count=None)
        if o == "/":
            return self._compile_operands(o, e, expect_count=2)
        self._error("operator={} not supported", o)

    def _compile_op_binary(self, op, operands):
        if len(operands) == 1:
            self._error("op={} requires two distinct operands={}", op, operands)
        l = operands[0]
        r = operands[1]
        o = _Operand(
            content=f"({l.content}{op}{r.content})",
            count=1,
            value=_OP_BINARY[op](l.value, r.value),
            is_decimal=True,
        )
        # TODO: if the fmt is not the same, that may be ok, because no fmt (decimal)
        #      for divide, do you have a format, e.g. $/$ has no format by default.
        self._compile_op_options(o, operands)
        return o

    def _compile_op_multi(self, op, operands):
        r = _Operand(count=1, is_decimal=True)
        c = ""
        m = _OP_MULTI[op]
        v = decimal.Decimal(m.init)
        for o in operands:
            if len(c):
                c += ","
            c += o.content
            if "cells" in o:
                for p in o.cells:
                    v = m.func(v, p.value)
            else:
                v = m.func(v, o.value)
        self._compile_op_options(r, operands)
        return r.pkupdate(
            content=f"{m.xl}({c})",
            value=v,
        )

    def _compile_op_options(self, res, operands):
        for o in operands:
            for x in "fmt", "round_digits":
                if o.get(x) is None:
                    continue
                if x not in res:
                    res[x] = o[x]
                # TODO doc that grabs the leftmost fmt or round
                elif res[x] is not None and res[x] != o[x]:
                    res[x] = None
        return res

    def _compile_op_unary(self, op, operands):
        o = operands[0]
        return _Operand(
            content=f"({op}{o.content})",
            count=1,
            fmt=o.get("fmt"),
            is_decimal=True,
            round_digits=o.round_digits,
            value=_OP_UNARY[op](o.value),
        )

    def _compile_operands(self, op, operands, expect_count):
        n = 0
        z = []
        for o in operands:
            if not isinstance(o, (list, tuple)):
                o = [o]
            e = self._compile_expr(o)
            if not e.is_decimal:
                self._error("operand={} must numeric for op={}", e.value, op)
            z.append(e)
            n += e.count
        if expect_count is None:
            return self._compile_op_multi(op, z)
        if expect_count != n:
            x = (
                "; You might need to be more specific link names to avoid automatic link operand grouping."
                if max(expect_count, 2) < n
                else ""
            )
            self._error(
                "operator={} expect_count={} operands, not actual={} operands={}{x}",
                op,
                expect_count,
                n,
                operands,
            )
        if expect_count == 1:
            return self._compile_op_unary(op, z)
        if expect_count == 2:
            return self._compile_op_binary(op, z)
        raise AssertionError(f"incorrect expect_count={expect_count}")

    def _compile_ref(self, link, expect_count):
        def _xl_id(other):
            r = ""
            if other.sheet != self.sheet:
                r = f"'{other.sheet.title}'!"
            return r + other.xl_id

        l = self.workbook.links
        if link not in l:
            self._error("link={} not found", link)
        p = None
        n = 0
        z = []
        for c in sorted(l[link], key=lambda x: x.sort_index):
            c._compile_pass2()
            n += 1
            # _ROW_MODULUS ensures a gap so columns are separated by more than "+1"
            # and this will never link the wrong column
            if (
                p is not None
                and p.sheet == c.sheet
                and p.sort_index + 1 == c.sort_index
            ):
                p = z[-1][1] = c
                continue
            z.append([c, None])
            p = c
        r = ""
        for x in z:
            if len(r):
                r += ","
            r += _xl_id(x[0])
            if x[1] is not None:
                r += ":" + _xl_id(x[1])
        if expect_count is not None and expect_count != n:
            self._error(
                "incorrect operands={} expect={} count={}",
                l[link],
                expect_count,
                n,
            )
        # _assert_link_pair validates that the cells are the same type
        return _Operand(
            cells=l[link],
            content=r,
            count=n,
            fmt=p.get("fmt"),
            # TODO(robnagler)
            # default round_digits: if p has round_digits explicit, does not
            # matter, because target round digits overrides, always. format
            # is different, though.
            round_digits=p.round_digits,
            value=p.value if n == 1 else None,
            is_decimal=p.is_decimal,
        )

    def _compile_str(self, value):
        self.round_digits = None
        self.value = self.content = value
        self.is_decimal = False

    def _save(self):
        f = _Fmt(self)
        self.sheet.width(_XL_COLS[self.col_num], f.width(self))
        f = self.workbook.xl_fmt(f)
        if self._is_expr:
            self.sheet.xl.write_formula(
                self.xl_id, formula=self.content, value=self.value, cell_format=f
            )
        else:
            self.sheet.xl.write(self.xl_id, self.content, f)
        if self.is_decimal:
            self.sheet.text_cells.append(self.xl_id)


class _Fmt(PKDict):
    _MAP = PKDict(
        top=PKDict(top=True),
        bold=PKDict(bold=True),
        currency="$#,##0.0",
        decimal="0.0",
        percent="0.0%",
        text="@",
    )
    _SPECIAL = re.compile("[$%]")
    _WIDTH_SLOP = 1

    def width(self, cell):
        if not cell.is_decimal:
            return len(cell.value)
        n = self._WIDTH_SLOP
        x = cell.value.as_tuple()
        i = len(x.digits) + x.exponent
        n += i
        if "," in self.num_format:
            n += i // 3
        n += cell.round_digits
        if n and self._SPECIAL.search(self.num_format):
            n += 1
        return n

    def __init__(self, cell):
        def _num(fmt):
            if cell.round_digits is None:
                return fmt
            if cell.round_digits == 0:
                x = ""
            else:
                x = "." + "0" * cell.round_digits
            fmt = fmt.replace(".0", x)
            return fmt

        for k in (
            "font",
            "border",
        ):
            if k in cell.defaults:
                self.update(self._MAP[cell.defaults[k]])
        for x in _SPACES.split(cell.get("fmt", "")):
            if len(x) == 0:
                continue
            f = self._MAP[x]
            if isinstance(f, dict):
                self.update(f)
            else:
                self.num_format = _num(f)
        if "num_format" not in self:
            self.num_format = _num(
                self._MAP[cell.defaults["num_fmt" if cell.is_decimal else "str_fmt"]],
            )

    def __str__(self):
        return ";".join([f"{k}={self[k]}" for k in sorted(self)])


class _Operand(PKDict):
    pass


class _Row(_Base):
    def __init__(self, cfg):
        """Creates a row"""
        super().__init__(cfg)
        self.cells = PKDict()

    def add_cell(self, col, content_or_cell, **kwargs):
        """Adds a cell to `col` in `self`

        Args:
            col (str): name of column to add
            content_or_cell (object): See `_Base.cell` for arguments
        """
        if col in self.cells:
            self._error(
                "cell={} already exists in cells={}", col, sorted(self.cells.keys())
            )
        self.cells[col] = (
            self.cell(content_or_cell, **kwargs).pkupdate(col=col)._relations(self)
        )
        return self

    def add_cells(self, *args, **kwargs):
        """Adds `values` to the row/header/footer

        For the footer, the first col must match first header

        Args:
            values (dict or list, tuple): ordered col=label/cell, where col is a keyword name or list of cells
            kwargs (dict): ordered col=label/cell
        Returns:
            self: row/header/footer
        """
        if len(args) == 0:
            # Allow empty case for call to row(), header(), etc.
            c = PKDict() if len(kwargs) == 0 else kwargs
        elif len(args) > 1:
            raise self._error("too many (>1) args={}", args)
        else:
            if isinstance(args[0], dict):
                c = PKDict(args[0])
            elif isinstance(args[0], (tuple, list)):
                # Really only useful for headers case
                c = PKDict(zip(args[0], args[0]))
            c.update(kwargs)
        for k, v in c.items():
            self.add_cell(k, v)
        return self

    def _children(self):
        return self.cells.values()

    def _compile_pass1(self):
        for k, v in self.cells.items():
            if k not in self.parent.cols:
                self._error("column={} does not exist; cell={}", k, v)
        s = set()
        r = self.row_num
        for i, n in enumerate(self.parent.cols, _COL_NUM_1):
            if n not in self.cells:
                self._error("column={} not found", n)
            self.cells[n].pkupdate(
                col_num=i,
                row_num=r,
                sort_index=i * _ROW_MODULUS + r,
                xl_id=f"{_XL_COLS[i]}{r}",
            )._compile_pass1()

    def _save(self):
        for c in self._children():
            c._save()


class _Footer(_Row):
    def __init__(self, cfg):
        super().__init__(cfg)
        self.defaults.pksetdefault(border="top")


class _Header(_Row):
    def __init__(self, cfg):
        super().__init__(cfg)
        self.defaults.pksetdefault(font="bold")


class _Sheet(_Base):
    def __init__(self, cfg):
        """Appends Sheet to Workbook

        Args:
            cfg (dict): configuration, e.g. title, defaults
        """
        super().__init__(cfg)
        self.tables = []

    def table(self, **kwargs):
        """Appends table to sheets

        Args:
            title (str): debug label for the table
            defaults (PKDict): default values, e.g. round_digits
        """
        return self._child(self.tables, _Table, kwargs)

    def width(self, col, width):
        if col not in self._col_widths or self._col_widths[col] <= width:
            self._col_widths[col] = width

    def _children(self):
        return self.tables

    def _compile_pass1(self):
        r = _ROW_NUM_1
        for t in self._children():
            r = t._compile_pass1(r)

    def _save(self):
        self.xl = self.workbook.xl.add_worksheet(self.title)
        self.text_cells = []
        self._col_widths = PKDict()
        for c in self._children():
            c._save()
        for c, w in self._col_widths.items():
            self.xl.set_column(f"{c}:{c}", width=w)
        if self.text_cells:
            # sqref is represented as list of cells separated by spaces
            self.xl.ignore_errors({"number_stored_as_text": " ".join(self.text_cells)})
        self.xl = None
        self.text_cells = None
        self._col_widths = None


class _Table(_Base):
    def __init__(self, cfg):
        """Append a table to Sheet

        The first row (be it header, row, or footer) defines the
        column names and the maximum width of the table.

        The column names must match this first row. A cell content
        may be None, which is an empty row.

        Args:
            cfg (dict): configuration, e.g. defaults
        """
        super().__init__(cfg)
        self.headers = []
        self.rows = []
        self.footers = []

    def footer(self, *args, **kwargs):
        """Append a footer

        The first footer will be separated from the table with top border.

        Args:
            cells (dict, kwargs, list): ordered col=cell; coll must match first header
        """
        return self._child(self.footers, _Footer, args, kwargs)

    def header(self, *args, **kwargs):
        """Append a header

        The first header defines the column names and the width of the table.

        Args:
            cells (dict, kwargs, list): ordered col=label, where col is a keyword name
        """
        return self._child(self.headers, _Header, args, kwargs)

    def row(self, *args, **kwargs):
        """Append a row

        Args:
            cells (dict, kwargs, list): ordered col=label, where col is a keyword name
        """
        return self._child(self.rows, _Row, args, kwargs)

    def _child(self, children, child, args, kwargs):
        return super()._child(children, child, PKDict()).add_cells(*args, **kwargs)

    def _children(self):
        return self.headers + self.rows + self.footers

    def _compile_pass1(self, row_num):
        self.first_row_num = row_num
        for r in self._children():
            if "cols" not in self:
                self.cols = tuple(r.cells.keys())
                self.col_set = frozenset(self.cols)
            r.row_num = row_num
            r._compile_pass1()
            row_num += 1
        return row_num

    def _save(self):
        for c in self._children():
            c._save()


def _init():
    global _XL_COLS, _DIGITS_TO_PLACES
    if _XL_COLS:
        return
    # really, you can 16384 columns, but we only support 702 (26 + 26*26)
    x = [chr(ord("A") + i) for i in range(26)]
    v = ([None] * _ROW_NUM_1) + x.copy()
    for c in x:
        v.extend([c + d for d in x])
    _XL_COLS = tuple(v)
    x = ("1",) + tuple(
        (("." + ("0" * i) + "1") for i in range(16)),
    )
    _DIGITS_TO_PLACES = tuple((decimal.Decimal(d) for d in x))


def _rnd(v, digits):
    return v.quantize(
        _DIGITS_TO_PLACES[digits],
        rounding=decimal.ROUND_HALF_UP,
    )


_init()



# -*- coding: utf-8 -*-
"""pytest for `pykern.pkdebug`

:copyright: Copyright (c) 2015-2020 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function

import inspect
import os
import os.path
import pytest
import re
import six

# Do not import anything from pykern so config can be fresh


@pytest.fixture(autouse=True)
def setup():
    # Test without logging redirects, because need to test native and then
    # test _logging_uninstall(). Need to clear any output or controls
    from pykern import pkdebug

    pkdebug.cfg.output = None
    pkdebug.cfg.control = None
    pkdebug.cfg.redirect_logging = False
    pkdebug.cfg.want_pid_time = False
    pkdebug.init()


def test_init(capsys):
    from pykern import pkunit

    f = pkunit.empty_work_dir().join("f1")
    from pykern.pkdebug import pkdp, init

    init(output=f)
    pkdp("init1")
    out, err = capsys.readouterr()
    assert "" == err, "When output is a file name, nothing goes to err"
    from pykern import pkio

    assert "init1\n" in pkio.read_text(f), "File output should contain msg"
    init(output=None, want_pid_time=True)
    pkdp("init2")
    out, err = capsys.readouterr()
    assert re.search(
        r"\w{3} .\d \d\d:\d\d:\d\d +\d+ +\d+ ", err
    ), "When output has time, matches regex"


def test_init_deviance(capsys):
    """Test init exceptions"""
    import pykern.pkdebug as d

    output = six.StringIO()
    d.init(control=r"(regex is missing closing parent", output=output)
    out, err = capsys.readouterr()
    assert not d._have_control, "When control re.compile fails, _printer is not set"
    assert (
        "compile error" in output.getvalue()
    ), "When an exception in init(), output indicates init failure"
    assert err == "", "When an exception in init() and output, stderr is empty"
    d.init(control=r"[invalid regex", output="/invalid/file/path")
    assert (
        not d._have_control
    ), "When invalid control regex, _have_control should be false"
    out, err = capsys.readouterr()
    assert (
        "compile error" in err
    ), "When exception in init() and output invalid, init failure written to stderr"


def test_ipython():
    try:
        import IPython
    except ImportError:
        pytest.skip("ipython not available; install manually")
    import pykern.pkdebug
    from pykern.pkdebug import pkdp
    from pykern import pkcompat

    pykern.pkdebug.init(output=None)
    # Overwrite the _ipython_write method. This doesn't test how ipython is
    # running. We'll do that separately
    save = []

    def _write(msg):
        save.append(msg)

    try:
        pykern.pkdebug._ipython_write = _write
        pkdp("abcdefgh")
        assert (
            "abcdefgh" in save[0]
        ), "When _ipython_write is set, should be called if no output"
    finally:
        pykern.pkdebug._ipython_write = None
    import subprocess

    try:
        p = subprocess.Popen(
            [
                "ipython",
                "--colors",
                "NoColor",
                "-c",
                'from pykern.pkdebug import pkdp; pkdp("abcdef")',
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        # Not a brilliant test, but does demonstrate that write_err works
        assert "<module> abcdef" in pkcompat.from_bytes(
            p.stderr.read()
        ), "When in IPython, pkdp() should output to stdout"
        # We make this rigid, because we want to know when IPython interpreter changes
        o = pkcompat.from_bytes(p.stdout.read())
        assert re.search(
            "Out\\[1\\]: \n?'abcdef'", o
        ), "IPython pkdp() is evaluated and written to stdout {}".format(o)

    except OSError as e:
        # If we don't have IPython, then ignore error
        import errno

        if e.errno != errno.ENOENT:
            reraise


def test_logging(capsys, caplog):
    """Verify basic output"""
    import logging
    from pykern import pkdebug

    logging.warn("warn_xyzzy")
    # py.test captures logging by default
    err = caplog.text
    caplog.clear()
    assert (
        "%(levelname)s:%(name)s:%(message)s" == logging.BASIC_FORMAT
    ), "logging.BASIC_FORMAT is not what is assumed in pkdebug and tests below"
    assert "warn_xyzzy" in err, "When logging is first initialized, warn should output"
    logging.info("info_xyzzy")
    out, err = capsys.readouterr()
    assert "" == err, "When logging is first initialized, info should not output"
    pkdebug.init(control=None, output=None, redirect_logging=True)
    logging.debug("debug_xyzzy")
    l = logging.getLogger("whatever")
    l.debug("whatever_debug_xyzzy")
    out, err = capsys.readouterr()
    assert "" == err, "When no control, logging.debug should not output anything"
    logging.info("info_xyzzy")
    l.info("whatever_info_xyzzy")
    out, err = capsys.readouterr()
    r = re.compile(
        r"INFO:root:info_xyzzy.*?INFO:whatever:whatever_info_xyzzy",
        flags=re.DOTALL,
    )
    matches = r.findall(err)
    assert 1 == len(matches), (
        "When no control, logging.info should output once per log; err=" + err
    )
    logging.warn("warn_xyzzy")
    l.warn("whatever_warn_xyzzy")
    out, err = capsys.readouterr()
    r = re.compile(
        r"WARNING:root:warn_xyzzy.*?WARNING:whatever:whatever_warn_xyzzy",
        flags=re.DOTALL,
    )
    matches = r.findall(err)
    assert 1 == len(matches), (
        "When no control, logging.warn should output once per log; err=" + err
    )
    try:
        1 / 0
    except ZeroDivisionError as e:
        logging.exception("zde")
    out, err = capsys.readouterr()
    assert re.search("Traceback.*ZeroDiv", err, flags=re.DOTALL)
    pkdebug.init(control="xyzzy", output=None, redirect_logging=True)
    logging.debug("debug_xyzzy")
    l.debug("whatever_debug_xyzzy")
    out, err = capsys.readouterr()
    assert (
        "DEBUG:root:debug_xyzzy" in err
    ), "When have control that matches, logging.debug should output"
    assert (
        "DEBUG:whatever:whatever_debug_xyzzy" in err
    ), "When have control that matches, logging.debug should output"
    pkdebug.init(control="xyzzy", output=None, redirect_logging=False)
    # Assumes default logging level is WARN
    logging.debug("dbg_xyzzy")
    l.debug("whatever_debug_xyzzy")
    logging.info("info_xyzzy")
    l.info("whatever_info_xyzzy")
    logging.warn("warn_xyzzy")
    # py.test captures logging by default
    err = caplog.text
    caplog.clear()
    assert (
        "warn_xyzzy" in err and "dbg" not in err and "info" not in err
    ), "When logging is not redirected, info and debug should not output"


def test_pkdc(capsys):
    """Verify basic output"""
    # The pkdc statement is four lines forward, hence +4
    this_file = os.path.relpath(__file__)
    control = (
        this_file + ":" + str(inspect.currentframe().f_lineno + 6) + ":test_pkdc t1"
    )
    from pykern import pkdebug
    from pykern.pkdebug import pkdc, init, pkdp

    init(control=control)
    pkdc("t1")
    out, err = capsys.readouterr()
    assert (
        control + "\n" == err
    ), "When control exactly matches file:line:func msg, output is same"
    pkdc("t2")
    out, err = capsys.readouterr()
    assert "" == err, "When pkdc msg does not match control, no output"
    init(control="t3")
    pkdc("t3 {}", "p3")
    out, err = capsys.readouterr()
    assert "test_pkdc t3" in err, "When control is simple msg match, expect output"
    assert (
        "t3 p3\n" in err
    ), "When positional format *args, expect positional param in output"
    output = six.StringIO()
    init(control="t4", output=output)
    pkdc("t4 {k4}", k4="v4")
    out, err = capsys.readouterr()
    assert (
        "test_pkdc t4 v4" in output.getvalue()
    ), "When params is **kwargs, value is formatted from params"
    assert "" == err, "When output is passed to init(), stderr is empty"


def test_pkdc_deviance(capsys):
    """Test max exceptions"""
    import pykern.pkdebug as d

    d.init(control=".")
    for i in range(d.MAX_EXCEPTION_COUNT):
        d.pkdc("missing format value {}")
        out, err = capsys.readouterr()
        assert (
            "invalid format" in err
        ), "When fmt is incorrect, output indicates format error"
    d.pkdc("any error{}")
    out, err = capsys.readouterr()
    assert "" == err, "When exception_count exceeds MAX_EXCEPTION_COUNT, no output"


def test_pkdexc():
    """Verify `pkdexc` includes down the stack that print_exception"""
    from pykern.pkdebug import init, pkdexc

    init()

    def force_error():
        xyzzy

    def tag1234():
        try:
            force_error()
        except:
            return pkdexc()

    actual = tag1234()
    for expect in "xyzzy", "force_error", "tag1234", "test_pkdexc":
        assert expect in actual, "{}: call not found: {}".format(expect, actual)
    assert not re.search(
        r"tag1234.*tag1234.*tag1234", actual, flags=re.DOTALL
    ), "tag1234: found routine thrice in exception stack: {}".format(actual)


def test_pkdp(capsys):
    """Basic output and return with `pkdp`"""
    from pykern.pkdebug import pkdp, init

    init()

    def _should_see():
        assert 333 == pkdp(333)

    _should_see()
    out, err = capsys.readouterr()
    assert str(333) in err, "When pkdp called, arg should be converted to str,"
    assert "should_see" in err

    def _should_not_see(msg):
        pkdp(msg, pkdebug_frame=inspect.currentframe().f_back)

    _should_not_see("my msg")
    out, err = capsys.readouterr()
    assert "my msg" in err
    assert "should_not_see" not in err


def test_pkdpretty():
    """Pretty printing arbitrary objects`"""
    from pykern.pkdebug import pkdpretty
    import six

    recursive = []
    any_obj = object()
    recursive.append(recursive)
    for obj, expect in (
        ('{"a":1}', '{\n    "a": 1\n}\n'),
        ('{"a":1}', '{\n    "a": 1\n}\n'),
        (
            {"b": set([1])},
            "{   'b': set([1])}\n" if six.PY2 else "{'b': {1}}\n",
        ),
        (recursive, recursive),
        (any_obj, any_obj),
    ):
        assert expect == pkdpretty(obj)


def _z(msg):
    """Useful for debugging this module"""
    with open("/dev/tty", "w") as f:
        f.write(str(msg) + "\n")



# -*- coding: utf-8 -*-
"""test github_orgmode

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import re
import pytest


def test_issues(monkeypatch):
    from pykern import pkunit, pkjson, pkio
    from pykern.pkcli import github_orgmode
    from pykern.pkcli import github
    from pykern.pkdebug import pkdp
    from pykern.pkcollections import PKDict

    class _MockGitHub(github.GitHub):
        def repo_arg(self, repo):
            return _MockRepoOrGitHub3(repo)

        def login(self):
            return _MockRepoOrGitHub3("")

    class _MockIssue(PKDict):
        def as_dict(self):
            return self

        def edit(self, **kwargs):
            pass

        @property
        def issue(self):
            """has to be a property to avoid recursion on canonicalize"""
            return self

        def pull_request(self, **kwargs):
            return self.get("pull_request_urls")

    class _MockRepoOrGitHub3(PKDict):
        def __init__(self, repo):
            self._case = repo
            self.pkupdate(pkjson.load_any(pkio.read_text("repo.json")))

        def issues(self, *args, **kwargs):
            return [_MockIssue(v) for v in self._issues]

        def labels(self, *args, **kwargs):
            return self._labels

        def me(self):
            # it's only used for the output file
            return PKDict(login="assignee_issues")

        def milestones(self, *args, **kwargs):
            return self._milestones

        def search_issues(self, *args, **kwargs):
            return [_MockIssue(v) for v in self._issues]

    monkeypatch.setattr(github, "GitHub", _MockGitHub)
    for d in pkunit.case_dirs():
        m = re.sub("-.+", "", d.purebasename)
        # repo name is ignored, just used for debugging
        if m == "from_issues":
            a = github_orgmode.from_issues(d.purebasename, org_d=d)
        elif m == "to_issues":
            a = github_orgmode.to_issues(org_path=d.join(f"{d.purebasename}.org"))
        elif m == "assignee_issues":
            a = github_orgmode.assignee_issues(user="assignee_issues", org_d=d)
        else:
            pkunit.pkfail("case={} unknown method", d.purebasename)
        pkjson.dump_pretty(a, filename="res.json")



# -*- coding: utf-8 -*-
"""Backwards and forward compatible Python utilities

Functions here will be available forever, even when functions are
removed from Python.

Some functions are no longer necessary, e.g. `unicode_getcwd`. Others
are still convenient, e.g. `from_bytes` is useful when you don't know
whether it is a `str` or not.

:copyright: Copyright (c) 2015-2023 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html

"""
# Limit pykern imports so avoid dependency issues for pkconfig
import inspect
import itertools
import locale
import os
import subprocess
import sys


def from_bytes(value):
    """Converts value to a str

    If `value` is not a str, decode with utf-8.

    If already str, does nothing.

    Args:
        value (object): The string or object to be decoded.

    Returns:
        bytes: encoded string
    """
    if isinstance(value, str):
        return value
    _assert_type(value, (bytes, bytearray))
    return value.decode("utf-8")


def locale_str(value):
    """Converts a value to a unicode str unless already unicode.

    Args:
        value (object): The string or object to be decoded, may be None.

    Returns:
        str: decoded string (PY2: type unicode)
    """
    if value is None:
        return None
    if isinstance(value, (bytes, bytearray)):
        return value.decode(locale.getpreferredencoding())
    _assert_type(value, str)
    return value


def to_bytes(value):
    """Converts a value to bytes

    If `value` is a str, encode with utf-8.

    If already bytes, does nothing.

    Args:
        value (object): The string or object to be encoded.

    Returns:
        bytes: encoded string
    """
    if isinstance(value, bytes):
        return value
    _assert_type(value, str)
    return bytes(value, "utf-8")


def unicode_getcwd():
    """:func:`os.getcwd` unicode wrapper

    Returns:
        str: current directory
    """
    return os.getcwd


def unicode_unescape(value):
    """Convert escaped unicode and Python backslash values in str

    Args:
        value (str): contains escaped characters
    Returns:
        str: unescaped string
    """
    return value.encode("utf-8").decode("unicode-escape")


def zip_strict(*iterables):
    """`zip` where iterables must be exact.

    ``strict`` option was added in 3.10. This function is always strict.
    Args:
        iterables (object): things to be zipped
    Returns:
        object: Iterator for zipping
    """
    if _version_ok(3, 10):
        return zip(*iterables, strict=True)
    return _zip_strict(iterables)


def _assert_type(value, typ):
    if not isinstance(value, typ):
        raise TypeError(
            '"value={:.20}<SNIP>" is not a {}, actual type={}'.format(
                repr(value),
                typ,
                type(value),
            ),
        )


def _version_ok(major, minor):
    return sys.version_info >= (major, minor)


def _zip_strict(iterables):
    """Code from https://stackoverflow.com/a/69485272"""

    def _first_tail():
        nonlocal first_stopped
        first_stopped = True
        return
        yield

    def _zip_tail():
        if not first_stopped:
            raise ValueError("first iterable is longer than rest")
        for x in itertools.chain.from_iterable(rest):
            raise ValueError(f"left over data={str(x):.100} in iterable")
            yield

    if len(iterables) < 2:
        return zip(*iterables)
    first_stopped = False
    iterables = iter(iterables)
    first = itertools.chain(next(iterables), _first_tail())
    rest = list(map(iter, iterables))
    return itertools.chain(zip(first, *rest), _zip_tail())



# -*- coding: utf-8 -*-
"""pytest for `pykern.resource`

:copyright: Copyright (c) 2015 Bivio Software, Inc.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest


def test_render_resource():
    import glob
    import os.path
    from pykern import pkio
    from pykern import pkjinja
    from pykern import pkunit

    t1 = pkunit.import_module_from_data_dir("t1")
    with pkunit.save_chdir_work():
        out = "out"
        expect = "\n!v1!\n"
        assert expect == t1.render(
            None
        ), "render_resource should return rendered template"
        assert not glob.glob("*"), "render_resource should not create any files"
        assert expect == t1.render(
            out
        ), "render_resource should return string even when writing to file"
        assert expect == pkio.read_text(
            out
        ), "With out, render_resource should write file"



# -*- coding: utf-8 -*-
"""test ci

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from pykern.pkdebug import pkdp
import re
import pytest


def test_check_prints_dot_prefix():
    from pykern.pkcli import ci, test
    from pykern import pkunit

    for d in pkunit.case_dirs():
        with pkunit.ExceptToFile():
            ci.check_prints()



# -*- coding: utf-8 -*-
"""Wrapper for Python formatter (currently, ``black``) to update and to validate a repository.

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from pykern.pkdebug import pkdp, pkdlog
import py
import pykern.pksubprocess


def run(*paths):
    """Run black formatter on `path`

    Args:
        *paths (strs or py.paths): strings or py.paths to file or directory
    """
    _black(paths)


def diff(*paths):
    """Run diff on file comparing formatted vs. current file state

    Args:
        *paths (strs or py.paths): strings or py.paths to file or directory
    """
    _black(paths, "--diff", "--check", "--no-color")


def check(*paths):
    """Returns True if there would be diff else return False

    Args:
        *paths (strs or py.paths): strings or py.paths to file or directory
    """
    try:
        _black(paths, "--check")
    except RuntimeError as e:
        if str(e) == "error exit(1)":
            pykern.pkcli.command_error(
                "paths={} need to be formatted", tuple(str(p) for p in paths)
            )
        raise


def _black(paths, *args):
    """Helper function invokes black with options

    Args:
         *paths (strs or py.paths): strings or py.paths to file or directory
         *args (strs): options to be passed to black
    """
    from pykern import pkunit
    from pykern.pkcli import test
    from pykern import pksetup

    pykern.pksubprocess.check_call_with_signals(
        [
            "black",
            "--quiet",
            "--extend-exclude",
            f"/{test.SUITE_D}/.*{pkunit.DATA_DIR_SUFFIX}/|/{pksetup.PACKAGE_DATA}/",
            *args,
            *paths,
        ],
    )



# -*- coding: utf-8 -*-
"""Demonstrate how to write reusable command line tools.

A `pykern.pkcli` module is designed to be used from any context
any other piece of code. Since we use `argh`, it makes it easy
to dispatch with a bit of glue so that all you have to do is write
a function.

Documentation should be provided as a normal docstring for the function.
See `echo`.

You should set up your project directory with `pykern.pkcli.projex`,
which will give you a shell command that gets automatically
installed. Each project has one such command. For example,
`pykern.pykern_console` is the module that invokes all `pykern`
pkcli functions.

To invoke `echo`, use the shell command::

    $ pykern pkexample echo something
    howdy: something

The output of the function is printed on stdout so if you don't
have to print messages, and the function can be used in other
contexts where stdout is not useful (e.g. returned to a brower).

If you invoke this module without arguments, you get a list of
functions you can call:

    $ pykern pkexample
    usage: pykern pkexample [-h] {echo,primes} ...
    pykern pkexample: error: too few arguments

This list only includes "public" functions (not beginning with
an underscore).

In your project, you'll have your own function, for example,
you can start Sirepo with::

    $ sirepo service http
     * Running on http://0.0.0.0:8000/
     * Restarting with reloader

The module `sirepo.sirepo_console` is automatically installed
as a shell command by ``setup.py``, and `pykern.pkcli` searches
for modules to invoke in the package `sirepo.pkcli`. In fact,
you can list out all modules by invoking ``sirepo`` without
any arguments::

    $ sirepo
    usage: sirepo module command [args...]
    Modules:
    celery
    elegant
    service
    srw
    warp

:copyright: Copyright (c) 2016 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from pykern import pkcli
import itertools


def echo(suffix, prefix="howdy: "):
    """Concatenate prefix and suffix

    Args:
        to_echo (str): what to print and must be at least five chars
        prefix (str): what to put in front of `to_echo` ["howdy: "]

    Returns:
        str: prefix + suffix
    """
    if len(suffix) < 5:
        # We raise argument and other errors with command_error, which
        # raises `pkcli.CommandError` in a nice way. We use Unix-style
        # error messages, that is, problematic object followed by colon
        # followed by an error message and any other details.
        pkcli.command_error("{}: suffix is too short (< 5 chars)", suffix)
    # Instead of printing messages in the simple case, we just
    # return the value. This way the function can be used in
    # other contexts. argh prints the return of the function to stdout.
    return prefix + suffix



# -*- coding: utf-8 -*-
"""Wrapper for :mod:`array` to simplify and make future compatible.

Not a complete wrapper. New routines added as required.

:copyright: Copyright (c) 2015 Bivio Software, Inc.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
from future.utils import bytes_to_native_str

import array

#: Future-proof typecode for double
DOUBLE_TYPECODE = bytes_to_native_str(b"d")

#: Future-proof typecode for float
FLOAT_TYPECODE = bytes_to_native_str(b"f")


def new_double(*args, **kwargs):
    """Creates a new double ("d") array

    Args are the same as :func:`array.array` except for typecode,
    which is passed by this module.

    Returns:
        array.array: New, initialized array
    """
    return array.array(DOUBLE_TYPECODE, *args, **kwargs)


def new_float(*args, **kwargs):
    """Creates a new float ("f") array

    Args are the same as :func:`array.array` except for typecode,
    which is passed by this module.

    Returns:
        array.array: New, initialized array
    """
    return array.array(FLOAT_TYPECODE, *args, **kwargs)



def foo():
    from pykern.pkdebug import pkdp
    #print('this should not set off check_prints')
    #   print('this should not be found either')
    #pkdp('neither should this')
    mention = (pkdp, print)



# -*- coding: utf-8 -*-
"""Continuous integration (CI) support

To execute all checks and tests in a CI script, use::

    pykern ci run

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from pykern import pkcli
from pykern import pkio
from pykern import pksetup
from pykern import pkunit
from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdp, pkdlog
import re

#: static/js/ext is where Sirepo stores 3rd party library files
_CHECK_FILES = PKDict(
    check_eof_newline=PKDict(
        exclude_files=re.compile(r"/static/js/ext/|node_modules/|^run/|^tests/|^venv/"),
        include_files=re.compile(r"\.(html|jinja|js|json|md|py|tsx|yml)$"),
    ),
    check_main=PKDict(
        exclude_files=re.compile(r".*(_console\.py)|^tests/|^venv/|^pyproject.toml$"),
        include_files=re.compile(r".*(\.py)$"),
    ),
    check_prints=PKDict(
        exclude_files=re.compile(
            rf".*(?:{pkunit.DATA_DIR_SUFFIX}|{pkunit.WORK_DIR_SUFFIX})/"
            + rf"|^\w+/{pksetup.PACKAGE_DATA}/"
            + r"|pkdebug[^/]*\.py$"
            + r"|(?:^|/)\."
            + r"|^run/"
            + r"|^venv/"
            + r"|^pyproject.toml$"
        ),
        include_files=re.compile(r".py$"),
    ),
)

_MAIN = re.compile(r"^if.*__name__")
_PRINT = re.compile(r"(?:\s|^)(?:pkdp|print)\(")
_PRINT_OK = re.compile(r"^\s*#\s*(?:pkdp|print)\(")


def check_eof_newline():
    """Recursively check repo for files missing newline at end of file.

    Checks html, jinja, js, json, md, py, tsx, and yml files.
    Excludes external files, tests, and run directory.
    """

    def _c(lines):
        return [] if lines[-1] == "" else [""]

    _check_files("check_eof_newline", _c)


def check_main():
    """Recursively check repo for modules with main programs.

    Checks .py files.
    Excludes ``<project>_console.py``, tests, and venv.
    """

    def _c(lines):
        r = []
        for j, l in enumerate(lines, start=1):
            if re.search(_MAIN, l):
                r.append(f":{j} {l}")
        return r

    _check_files("check_main", _c)


def check_prints():
    """Recursively check repo for (naked) print and pkdp calls.

    Checks .py files, excluding hidden files, pkdebug module, test data/work, run directory,
    package_data and venv.

    See the
    `DesignHints <https://github.com/radiasoft/pykern/wiki/DesignHints#output-for-programmers-logging>_`
    wiki for an explanation of why this check is necessary.

    If you really need a print, use `pykern.pkconst.builtin_print`.
    """

    def _c(lines):
        r = []
        for j, l in enumerate(lines, start=1):
            if re.search(_PRINT, l) and not re.search(_PRINT_OK, l):
                r.append(f":{j} {l}")
        return r

    _check_files("check_prints", _c)


def run():
    """Run the continuous integration checks and tests:

    #. Runs `check_prints`
    #. Checks formatting
    #. Runs `pykern.pkcli.test.default_command`
    """
    from pykern.pkcli import fmt, test
    from pykern import pkio

    check_eof_newline()
    check_main()
    check_prints()
    pkdlog(
        "Checking fmt diff. If a diff is printed below, you can fix this failure by running `pykern fmt run .` from the repo root."
    )
    fmt.diff(*_paths(pkio.py_path()))
    test.default_command()


def _check_files(case, check_file):
    def _error(m):
        pkcli.command_error("{}: {}", case, m)

    d = _CHECK_FILES[case]
    r = []
    n = 0

    c = pkio.py_path()
    for p in _paths(c):
        for f in pkio.walk_tree(p, d.include_files):
            f = c.bestrelpath(f)
            if re.search(d.exclude_files, f):
                continue
            n += 1
            for l in check_file(pkio.read_text(f).split("\n")):
                r.append(f"{f}{l}")

    if n == 0:
        _error("no files found")
    if r:
        _error("\n".join(r))


def _paths(cwd):
    def _exists(*names):
        return list(filter(lambda n: cwd.join(n).isfile(), names))

    if _exists("README.rst", "README.md") and (
        (x := _exists("setup.py")) or _exists("pyproject.toml")
    ):
        return (
            # POSIT: repo name
            cwd.basename,
            "tests",
            *x,
        )
    return (cwd,)



# -*- coding: utf-8 -*-
"""test fmt

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdp
import pytest


def test_check():
    from pykern.pkcli import fmt

    _cases_with_prefix_except("check", fmt.check)


def test_diff():
    from pykern.pkcli import fmt

    _cases_with_prefix_except("diff", fmt.diff)


def test_run():
    from pykern import pkunit
    from pykern.pkcli import fmt

    for d in pkunit.case_dirs("fmt_dir"):
        fmt.run(d)


def _cases_with_prefix_except(prefix, fn):
    from pykern import pkunit
    from pykern import pkio

    for d in pkunit.case_dirs(prefix):
        with pkunit.ExceptToFile():
            fn(d)



# -*- coding: utf-8 -*-
"""API wrapper

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdc, pkdlog, pkdp


class Args:
    # qargs
    pass


class API(PKDict):
    # qcall
    pass


class Result:
    # qres
    pass


class Spec:
    # qspec
    pass



# -*- coding: utf-8 -*-
"""pkconfig init test

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest


def test_init(pkconfig_setup):
    """Validate initializing a module"""
    pkconfig = pkconfig_setup(
        cfg=dict(
            P1_M1_BOOL3="",
            P1_M1_BOOL4="y",
            P1_M1_P6="2012-12-12T12:12:12Z",
        ),
        env=dict(P1_M1_REQ8="99"),
    )
    import dateutil.parser

    from p1.m1 import cfg

    assert "default2" == cfg["dict1"]["d2"], "Nothing should change dict1[d2]"
    assert (
        dateutil.parser.parse("2012-12-12T12:12:12Z") == cfg["p6"]
    ), "environ sets time value and m1._custom_p6"
    assert 999 == cfg.dynamic_default10, "When value is None, calls parser"
    assert False == cfg.bool1, "When bool1 is none, is False"
    assert True == cfg.bool2, "When bool2 is none, is True"
    pkconfig.reset_state_for_testing()
    assert False == cfg.bool3, "bool3 should be overriden to be False"
    assert True == cfg.bool4, "bool4 should be overriden to be True"



def force_arg_per_line_and_str(
    blah,
    this,
    be,
    a,
    very,
    poorly,
    formatted,
    function,
    that,
    noone,
    should,
    ever,
    use,
    it,
    has,
    way,
    too,
    many,
    arguments,
):
    """Messy Docstring
    If the element does not exist or is not indexable, fails silently with None.
    Args:
    dotted_key (str): what
    Returns:
    object: value of element or None
    """
    x = "Dw0kHyZ4DDIPMaLxZkRl2pFskoSzcDjSIU8czQA7H1PDXwhjQzDFY2lWeHmb1bKHSLEnlLf4bhTd"

    y = 100
    z = 200
    p, q = 1, 2

    foo(y + z + p + q)



"""pykern package

:copyright: Copyright (c) 2018 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

import importlib.metadata

try:
    __version__ = importlib.metadata.version("pykern")
except importlib.metadata.PackageNotFoundError:
    # We only have a version once the package is installed.
    pass



# -*- coding: utf-8 -*-
"""pytest for `pykern.pkconfig`

:copyright: Copyright (c) 2015-2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest


def test_channel_in(pkconfig_setup):
    """Validate channel_in()"""
    pkconfig = pkconfig_setup(
        env={
            "PYKERN_PKCONFIG_CHANNEL": "dev",
            "SOME_VALUE": "1",
            "some_value": "2",
        },
    )
    channel = "dev"
    not_channel = "alpha"
    assert pkconfig.channel_in(channel), "Should match configured channel"
    assert not pkconfig.channel_in(not_channel), "Should not match configured channel"
    assert pkconfig.channel_in(not_channel, channel), "Should match configured channel"
    with pytest.raises(AssertionError):
        pkconfig.channel_in("bad channel")


def test_flatten_values():
    from pykern.pkconfig import flatten_values
    from pykern import pkcollections

    base = pkcollections.Dict()
    flatten_values(base, {"aa": 1, "bb": {"cc": 3}})
    assert base["bb_cc"] == 3
    flatten_values(base, {"bb": {"dd": 4}})
    assert base["bb_cc"] == 3
    assert base["bb_dd"] == 4


def test_parse_bytes():
    from pykern.pkconfig import parse_bytes

    assert 10 == parse_bytes(10)
    assert 20 == parse_bytes("20")
    assert 1024 == parse_bytes("1k")
    assert 2097152 == parse_bytes("2m")
    assert 751619276800 == parse_bytes("700GB")
    assert 4398046511104 == parse_bytes("004Tb")


def test_parse_seconds():
    from pykern.pkconfig import parse_seconds, parse_secs

    assert 999 == parse_seconds("999")
    assert 3600 == parse_seconds("1:0:0")
    assert 90061 == parse_seconds("1d1:1:1")
    assert 86461 == parse_seconds("1d1:1")
    assert 86401 == parse_seconds("1d1")
    assert 172800 == parse_seconds("2d")
    # deprecated form
    assert 1 == parse_secs(1)



# -*- coding: utf-8 -*-
"""PyTest for :mod:`pykern.pkarray`

:copyright: Copyright (c) 2015 Bivio Software, Inc.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest


def test_new_double():
    from pykern import pkarray

    d = pkarray.new_double()
    assert 0 == len(d), "new_double without initializer should be empty"
    d = pkarray.new_double([3, 5])
    assert 2 == len(d), "new_double with initializer, should be non-zero"
    assert float(5) == d[1], "new_double should intitialize to a float"


def test_new_float():
    from pykern import pkarray

    d = pkarray.new_float()
    assert 0 == len(d), "new_float without initializer should be empty"
    d = pkarray.new_float([3, 5])
    assert 2 == len(d), "new_float with initializer, should be non-zero"
    assert float(5) == d[1], "new_float should intitialize to a float"



"""Support routines, including run dir resolution.

:copyright: Copyright (c) 2023 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

# Root module: Limit imports to avoid dependency issues
import os.path
import sys


_ACCEPTABLE_CONTROL_CODE_RATIO = 0.33
_DEFAULT_ROOT = "run"
_DEV_ONLY_FILES = ("setup.py", "pyproject.toml")
_VALID_ASCII_CONTROL_CODES = frozenset((0x7, 0x8, 0x9, 0xA, 0xB, 0xC, 0xD, 0x1B))


def cfg_absolute_dir(value):
    """Parser function for absolute dir config value

    Args:
        value (str): string absolute path to dir

    Returns:
        py.path.Local absolute path to dir
    """
    from pykern import pkio, pkconfig

    if not os.path.isabs(value):
        pkconfig.raise_error("must be absolute")
    if not os.path.isdir(value):
        pkconfig.raise_error("must be a directory and exist")
    return pkio.py_path(value)


def dev_run_dir(package_object):
    """Resolves run directory if in development mode and creates it if it doesn't exist.

    Note that this assumes a file structure like pykern's where <project_name>'s __init__.py
    lives in <project_name>/<project_name>/

    Args:
        package_object (object): A python object whose root package we are resolving the run directory for

    Returns:
        py.path.local: The absolute path to the run directory.

    Raises:
        AssertionError: Raised when not in development mode
    """
    from pykern import pkio, pkconfig, pkinspect

    def _check_dev_files(root):
        """Check for files that only exist in development"""
        return any([root.join(f).check() for f in _DEV_ONLY_FILES])

    if not pkconfig.in_dev_mode():
        raise AssertionError("run dir root must be configured except in dev")
    r = (
        pkio.py_path(
            sys.modules[pkinspect.root_package(package_object)].__file__,
        )
        .dirpath()
        .dirpath()
    )
    if not _check_dev_files(r):
        # Don't run from an install directory
        r = pkio.py_path()
    return pkio.mkdir_parent(r.join(_DEFAULT_ROOT))


def is_pure_text(value, is_truncated=False):
    """Guesses if value is text data using heuristics:

    Checks if value can be utf-8 decoded.

    If fails to decode and is_truncated, probes backwards up to 4 chars
    (4 is maximum length of a utf-8 char) in case a valid utf-8
    char was truncated at the boundary of value.

    Returns False if null byte is present.

    On successful decode, checks that the amount of control codes not
    typical of text data do not exceed one third of the total characters.

    Args:
        value (bytes): bytes data
        is_truncated (bool): whether or not value has been truncated

    Returns:
        bool: True if bytes_data is likely pure text, false if likely binary
    """

    def _is_accepted_control_code_ratio(text_value):
        n = 0
        for c in text_value:
            if ord(c) == 0:
                return False
            if ord(c) < 32 and ord(c) not in _VALID_ASCII_CONTROL_CODES:
                n += 1
        return (n / len(text_value)) < _ACCEPTABLE_CONTROL_CODE_RATIO

    def _try_utf8(chunk):
        try:
            return chunk.decode("utf-8", "strict")
        except UnicodeDecodeError:
            return False

    def _utf8_decoded(value):
        if not is_truncated:
            return _try_utf8(value)
        b = value[: len(value)]
        for _ in range(4):
            if d := _try_utf8(b):
                return d
            if len(b) <= 1:
                return False
            b = b[:-1]
        return False

    if value == b"":
        return True
    if d := _utf8_decoded(value):
        return _is_accepted_control_code_ratio(d)
    return False



# -*- coding: utf-8 -*-
"""Simplify rendering jinja2

:copyright: Copyright (c) 2015 Bivio Software, Inc.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
from pykern import pkinspect
from pykern import pkio
from pykern import pkresource
from pykern.pkdebug import pkdc, pkdp
import jinja2


#: Implicit extension including '.' added to resources
RESOURCE_SUFFIX = ".jinja"


def render_file(filename, j2_ctx, output=None, strict_undefined=False, jinja_env=None):
    """Render filename as template with j2_ctx.

    Args:
        basename (str): name without jinja extension
        j2_ctx (dict): how to replace values in Jinja2 template
        output (str): file name of output; if None, return str
        strict_undefined (bool): set `jinja2.StrictUndefined` if True
        jinja_env (dict): add values to jinja2 environment

    Returns:
        str: rendered template
    """
    t = pkio.read_text(filename)
    kw = dict(
        trim_blocks=True,
        lstrip_blocks=True,
        keep_trailing_newline=True,
        extensions=["jinja2.ext.do"],
    )
    if strict_undefined:
        kw["undefined"] = jinja2.StrictUndefined
    if jinja_env:
        kw.update(jinja_env)
    je = jinja2.Environment(**kw)
    res = je.from_string(t).render(j2_ctx)
    if output:
        pkio.write_text(output, res)
    return res


def render_resource(basename, *args, **kwargs):
    """Render a pkresource as a jinja template.

    Args:
        basename (str): name without `RESOURCE_SUFFIX`
        args (list): see func:`render_file` for rest of args and return
    """
    return render_file(
        pkresource.filename(
            basename + RESOURCE_SUFFIX,
            pkinspect.caller_module(),
        ),
        *args,
        **kwargs
    )



# -*- coding: utf-8 -*-
"""JSON wrapper

:copyright: Copyright (c) 2017-2023 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import json


#: how bytes are encoded
ENCODING = "utf-8"

#: MIME type
MIME_TYPE = "application/json"

#: Content-Type MIME header
CONTENT_TYPE = f'{MIME_TYPE}; charset="{ENCODING}"'

_JSON_INT_MAX = 2**53 - 1
_JSON_INT_MIN = -(2**53) + 1


class Encoder(json.JSONEncoder):
    def default(self, obj):
        # Return Python object, and JSONEncoder._iterencode will encode
        return str(obj)


def dump_bytes(obj, **kwargs):
    """Formats as json as bytes for network transfer

    Args:
        obj (object): any Python object
        kwargs (object): other arguments to `dump_pretty`
    Returns:
        bytes: (unsorted) formatted JSON
    """
    return dump_pretty(obj, pretty=False, **kwargs).encode(ENCODING)


def dump_pretty(obj, filename=None, pretty=True, **kwargs):
    """Formats as json as string

    If an object is not encoded by default, will call str() on the
    object.

    Args:
        obj (object): any Python object
        filename (str or py.path): where to write [None]
        pretty (bool): pretty print [True]
        kwargs (object): other arguments to `json.dumps`

    Returns:
        str: sorted and formatted JSON
    """
    if pretty:
        res = (
            json.dumps(
                obj,
                indent=4,
                separators=(",", ": "),
                sort_keys=True,
                cls=Encoder,
                **kwargs,
            )
            + "\n"
        )
    else:
        res = json.dumps(obj, separators=(",", ":"), cls=Encoder, **kwargs)
    if filename:
        from pykern import pkio

        pkio.write_text(filename, res)
    return res


def dump_str(obj, **kwargs):
    """Formats as a compact json as string

    Args:
        obj (object): any Python object
        kwargs (object): other arguments to `dump_pretty`
    Returns:
        str: (unsorted) formatted JSON
    """
    return dump_pretty(obj, pretty=False, **kwargs)


def load_any(obj, *args, **kwargs):
    """Parse object containing json into dict-like object.

    object_pairs_hook modifies the return type.

    Args:
        obj (object): str or object with "read" or py.path
        args (tuple): passed verbatim to json.loads()
        kwargs (dict): object_pairs_hook may be overriden

    Returns:
        object: parsed JSON
    """
    from pykern import pkcollections

    def _parse_int(value):
        if len(value) > 64:
            # 64 is pretty arbitrary, but reasonable in Python. Nothing in
            # JSON can be this large.
            raise ValueError(f"number={value:60s}... unreasonably large")
        if len(value) > 17:
            return float(value)
        res = int(value)
        if _JSON_INT_MIN <= res <= _JSON_INT_MAX:
            return res
        return float(res)

    kwargs.setdefault("parse_int", _parse_int)
    kwargs.setdefault("object_pairs_hook", pkcollections.object_pairs_hook)
    return json.loads(obj.read() if hasattr(obj, "read") else obj, *args, **kwargs)



# -*- coding: utf-8 -*-
"""PKDict abstraction and utils

`PKDict` is similar to :class:`argparse.Namespace`, but is a dict that allows
you to treat elements as attributes.

:copyright: Copyright (c) 2015-2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
# Limit pykern imports so avoid dependency issues for pkconfig
import copy
import collections.abc
import decimal
import types
import pykern.pkcompat


class PKDict(dict):
    """A subclass of dict that allows items to be read/written as attributes.

    The purpose of this is as a convenience in coding. You
    can refer to dictionary keys as attributes as long as they
    don't collide with the object's attributes. You should always
    use the `dict` interface to refer to the items of the dictionary
    programmatically, just like you would do with Javascript.

    You can reference any dict key as an attribute as long as it
    does not conflict with an attribute. For example, this works::

        x = PKDict()
        x.a = 1
        assert 1 == x.a
        x['a'] = 3
        assert 3 == x.a
        assert 'a' == x.keys()[0]

    You can't set an attribute that already exists. These calls throw
    exceptions::

        x.values = 1
        delattr(x, 'values')

    `dict` doesn't allow this anyway. However, you can't set or
    delete any existing attribute, even writable attributes. Indeed,
    you can't delete attributes at all. Subclasses should be "containers"
    only, not general objects.
    """

    def __delattr__(self, name):
        raise PKDictNameError("{}: you cannot delete attributes", name)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        # must match what CPython does exactly:
        # https://github.com/python/cpython/blob/d583738a87c3019dcfe06ed4a0002d1d6c9e9762/Objects/object.c#L899
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'"
        )

    def __setattr__(self, name, value):
        if name in dir(self):
            raise PKDictNameError(
                "{}: invalid key for PKDict matches existing attribute".format(name)
            )
        super(PKDict, self).__setitem__(name, value)

    def copy(self):
        """Override `dict.copy` to ensure the class of the return object is correct.

        Necessary because dict.copy will return a dict rather than PKDict.
        Calls `self.__class__(self)` which makes a shallow copy.
        Subclasses that do not accept this constructor form will need to override this method.

        Returns:
             object: shallow copy of self
        """
        return self.__class__(self)

    def pkdel(self, name, default=None):
        """Delete item if exists and return value

        The code will survive against concurrent access, but is not thread safe.

        Never throws KeyError.

        Args:
            name (object): item to delete
        Returns:
            object: value (if exists) or default
        """
        try:
            return self[name]
        except KeyError:
            return default
        finally:
            try:
                del self[name]
            except KeyError:
                pass

    def pkmerge(self, to_merge, make_copy=True):
        """Add `to_merge` to `self`

        Types are assumed to match and are not converted, e.g. dict is
        not converted to PKDict. Again, use `canonicalize` if you want
        to avoid type incompatibilities.

        `to_merge`'s values override `self`'s so if say, `to_merge` is ``{'x': None}``,
        then ``self.x`` will be `None` at the end of this call even if it had a value
        before.

        Lists from to_merge are prepended on this same principle, that
        is, to_merge "overrides" self, and prepending is defined as
        overriding. Lists must contain unique elements and duplicates will
        cause an error.

        This function recurses only on PKDicts.

        Args:
            to_merge (dict): elements will be copied into `self`
            make_copy (bool): deepcopy `to_merge` before merging [True]

        Returns:
            PKDict: self
        """

        def _type_err(key, base, merge):
            return AssertionError(
                f"key={key} type mismatch between (self) base={base} and to_merge={merge}"
            )

        if make_copy:
            to_merge = copy.deepcopy(to_merge)

        for k in list(to_merge.keys()):
            t = to_merge[k]
            s = self.get(k)
            if isinstance(s, dict) and isinstance(t, dict):
                s.pkmerge(t, make_copy=False)
            elif isinstance(s, list) and isinstance(t, list):
                # prepend the to_merge values (see docstring above)
                # NOTE: creates a new list
                self[k] = t + s
                # strings, numbers, etc. are hashable, but dicts and lists are not.
                # this test ensures we don't have dup entries in lists.
                y = [x for x in self[k] if isinstance(x, collections.abc.Hashable)]
                assert len(set(y)) == len(
                    y
                ), f"duplicates in key={k} list values={self[k]}"
            elif type(s) == type(t) or s is None or t is None:
                # Just replace, because t overrides type in case of None.
                # And if s is None, it doesn't matter.
                self[k] = t
            else:
                raise _type_err(k, s, t)
        return self

    def pknested_get(self, qualifiers):
        """Split key on dots or iterable and return nested get calls

        If `qualifiers` is a str, will split on dots. Otherwise, will be iterated.

        If an element is a list or tuple, tries to index as int.

        Throws KeyError if the dictionary key doesn't exist.

        Args:
            qualifiers (str or iterable): see above

        Returns:
            object: value of element
        """
        d = self
        for k in qualifiers.split(".") if isinstance(qualifiers, str) else qualifiers:
            try:
                d = d[k]
            except TypeError:
                try:
                    d = d[int(k)]
                    continue
                except (ValueError, TypeError):
                    pass
                raise
        return d

    def pknested_set(self, qualifiers, value):
        """Set nested location identified by `qualifiers` to `value`

        If `qualifiers` is a str, will split on dots. Otherwise, will be iterated.

        If an element does not exist, creates it as a PKDict.

        DOES NOT SUPPORT ints (lists) at this time.

        Args:
            qualifiers (str or iterable): see above
            value (any): assigned to qualifier's location in self

        Returns:
            object: self
        """
        q = qualifiers.split(".") if isinstance(qualifiers, str) else list(qualifiers)
        d = self
        for k in q[:-1]:
            if k not in d:
                d[k] = PKDict()
            d = d[k]
        d[q[-1]] = value
        return self

    def pksetdefault(self, *args, **kwargs):
        """Set values for keys if key not already in self

        Accepts args or kwargs, but not both.
        `args` must be an even number and are interpreted in (key, value) order.
        `kwargs` are accepted as key=value.

        If self does not have `key`, then it will be set to
        `value`. If `key` is already in self, its value is not changed.
        If `value` is a callable, it will be called, and
        `key` will be set to the returned value.

        Args:
            key (object): key that will be assigned `value` if not already set
            value (object): if callable, will be called, else verbatim
        Returns:
            object: self
        """
        if args and kwargs:
            raise AssertionError("one of args or kwargs must be set, but not both")
        if args:
            if len(args) % 2 != 0:
                raise AssertionError(
                    "args must be an even number (pairs of key, value)"
                )
            i = zip(args[0::2], args[1::2])
        else:
            i = kwargs.items()
        for k, v in i:
            self.__pksetdefault_one(k, v)
        return self

    def pksetdefault1(self, *args, **kwargs):
        """Get value if exists else set

        Accepts one key and one value, either as two positional args
        (``key, value``); or one keyword arg (``key=value``).


        If self does not have `key`, then it will be set to
        `value`. If `value` is a callable, it will be called, and
        `key` will be set to the returned value.

        Args:
            key (object): value to get or set
            value (object): if callable, will be called, else verbatim
        Returns:
            object: ``self[key]``; either `value` if just set, or preexisting value

        """
        if args and kwargs:
            raise AssertionError("one of args or kwargs must be set, but not both")
        if args:
            if len(args) == 2:
                return self.__pksetdefault_one(*args)
        elif len(kwargs) == 1:
            for k, v in kwargs.items():
                return self.__pksetdefault_one(k, v)
        raise AssertionError("must pass exactly one key and value")

    def pkunchecked_nested_get(self, qualifiers, default=None):
        """Split key on dots or iterable and return nested get calls

        If `qualifiers` is a str, will split on dots. Otherwise, will be iterated.

        If the element does not exist or is not indexable, fails silently and returns `default`.

        Throws KeyError if the dictionary key doesn't exist.

        Args:
            qualifiers (str or iterable):

        Returns:
            object: value of element
        """
        try:
            return self.pknested_get(qualifiers)
        except (KeyError, IndexError, TypeError, ValueError):
            return default

    def pkupdate(self, *args, **kwargs):
        """Call `dict.update` and return ``self``."""
        super(PKDict, self).update(*args, **kwargs)
        return self

    def __pksetdefault_one(self, key, value):
        if key not in self:
            self[key] = value() if callable(value) else value
        return self[key]


class PKDictNameError(NameError):
    """Raised when a key matches a builtin attribute in `dict`."""

    pass


def canonicalize(obj):
    """Convert to lists and PKDicts for simpler serialization

    Traverse `obj` to convert all values to forms that are compatible
    with serialization protocols like YAML or JSON.

    Simple objects are ensured to match their types e.g. bool, float,
    int, and str.  Objects that are instances of these, are converted
    to these to ensure they are basic types, that is,
    ``canonicalize(str_subclass('a'))`` will be converted to ``str('a')``.

    bytes and bytearrays will be converted to str.

    decimal.Decimal will converted to float.

    All objects are traversed. If no objects need to be converted,
    `obj` will be returned unmodified.

    Generators and iterables are converted to lists.

    Circularities are not detected so infinite recursion can occur.

    Args:
        obj (object): what to convert

    Returns:
        object: converted object (may or may not be the same)
    """
    o = obj
    if o is None:
        return o
    # Order matters so we don't convert bools to ints, since bools are ints.
    for x in (
        (bool,),
        (int,),
        (float,),
        (str,),
        (decimal.Decimal, float),
        ((bytes, bytearray), pykern.pkcompat.from_bytes),
        (
            dict,
            lambda y: PKDict({canonicalize(k): canonicalize(v) for k, v in y.items()}),
        ),
        (types.GeneratorType, lambda y: list(canonicalize(i) for i in y)),
        (collections.abc.Iterable, lambda y: list(canonicalize(i) for i in iter(y))),
    ):
        if isinstance(o, x[0]):
            return x[-1](o)
    raise ValueError(f"unable to canonicalize type={type(o)} value={repr(o):100}")


# Deprecated names
Dict = PKDict
DictNameError = PKDictNameError


def object_pairs_hook(*args, **kwargs):
    """Tries to use `PKDict` if PKDictNameError uses `dict`

    Useful for json, msgpack, etc.

    Returns:
        object: `PKDict` or `dict`
    """
    try:
        return PKDict(*args, **kwargs)
    except PKDictNameError:
        return dict(*args, **kwargs)


def unchecked_del(obj, *keys):
    """Deletes the keys from obj

    Args:
        obj (object): What to delete from (usually dict)
        keys (object): What to delete
    """
    for k in keys:
        try:
            del obj[k]
        except KeyError:
            pass



# -*- coding: utf-8 -*-
"""Run python code

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
# Avoid pykern imports so avoid dependency issues
import importlib.machinery
import importlib.util
import os.path
import sys


def run_path_as_module(fname):
    """Runs ``fname`` in a module

    Args:
        fname (str or py.path.local): file to be exec'd

    Returns:
        module: imported file as a module
    """
    fname = str(fname)
    mn = os.path.basename(fname).replace(".", "_")
    m = importlib.util.module_from_spec(importlib.machinery.ModuleSpec(mn, None))
    with open(fname, "rt") as f:
        code = compile(f.read(), fname, "exec")
    exec(code, m.__dict__)
    return m



"""pkcli.test test

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def setup_module(module):
    import os

    # Our CI (radiasoft/download/installers/python-ci sets these to 1,
    # but we do not want restarts or to stop out due to max failures
    os.environ.update(
        PYKERN_PKCLI_TEST_MAX_FAILURES="5",
        PYKERN_PKCLI_TEST_RESTARTABLE="0",
    )


def test_simple(capsys):
    from pykern import pkunit
    from pykern.pkcli import test

    with pkunit.save_chdir_work() as d:
        t = d.join("tests")
        pkunit.data_dir().join("tests").copy(t)
        with pkunit.pkexcept("FAILED=1 passed=2", "all tests"):
            test.default_command()
        o, e = capsys.readouterr()
        pkunit.pkre("1_test.py pass", o)
        pkunit.pkre("2_test.py FAIL", o)
        t.join("2_test.py").rename(t.join("2_test.py-"))
        # tests 1 & 3 together
        pkunit.pkre("passed=2", test.default_command())
        o, e = capsys.readouterr()
        pkunit.pkre("1_test.py pass", o)
        pkunit.pkre(
            "3_test.py pass\ntests/3_test.py::test_skip SKIPPED .always skips.",
            o,
        )
        # tests 1 and 3 individually
        pkunit.pkre("passed=1", test.default_command("tests/1_test.py"))
        o, e = capsys.readouterr()
        pkunit.pkre("1_test.py pass", o)
        pkunit.pkre("passed=1", test.default_command("tests/3_test.py"))
        o, e = capsys.readouterr()
        pkunit.pkre("3_test.py pass.*test_skip SKIPPED", o)
        # test 2 and 3 together
        t.join("2_test.py-").rename(t.join("2_test.py"))
        t.join("1_test.py").rename(t.join("1_test.py-"))
        with pkunit.pkexcept("FAILED=1 passed=1"):
            test.default_command()
        o, e = capsys.readouterr()
        pkunit.pkre("2_test.py FAIL", o)
        pkunit.pkre("x = 1 / 0", o)
        pkunit.pkre("test_skip SKIPPED", o)


def test_tests_dir(capsys):
    from pykern import pkio, pkdebug
    from pykern import pkunit
    from pykern.pkcli import test

    with pkunit.save_chdir_work() as d:
        pkunit.data_dir().join("tests").copy(d.join("tests"))
        with pkunit.pkexcept("FAILED=1 passed=2"):
            test.default_command()
        with pkunit.pkexcept("FAILED=1 passed=1"):
            test.default_command("skip_past=1_test")
        # to avoid confusing output with SKIPPED
        capsys.readouterr()



def force_arg_per_line_and_str(blah,this, be, a, very, poorly, formatted, function, that, noone, should, ever, use, it, has, way, too, many, arguments):
    '''Messy Docstring
If the element does not exist or is not indexable, fails silently with None.
Args:
dotted_key (str): what
Returns:
object: value of element or None
        '''
    x='Dw0kHyZ4DDIPMaLxZkRl2pFskoSzcDjSIU8czQA7H1PDXwhjQzDFY2lWeHmb1bKHSLEnlLf4bhTd'





    y= 100
    z= 200
    p, q = 1, 2


    foo(y+z+p+q)



# -*- coding: utf-8 -*-
"""restarts twice

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest


def test_twice():
    from pykern import pkunit
    from pykern import pkio
    from pykern import pkdebug

    p = pkio.py_path().join("twice")
    if p.exists():
        # success
        return
    p.write("test_twice")
    pkunit.restart_or_fail("restart infinitely")



# -*- coding: utf-8 -*-
"""pytest for `pykern.resource`

:copyright: Copyright (c) 2015 Bivio Software, Inc.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest


def test_filename():
    import importlib
    import os.path
    from pykern import pkunit
    from pykern import pkresource

    d = pkunit.data_dir()
    t1 = importlib.import_module(d.basename + ".t1")
    assert t1.somefile().startswith(
        "anything"
    ), 'When somefile is called, it should return the "anything" file'
    n = pkresource.filename("test.yml", pkresource)
    sn = [n]

    def _tail():
        (sn[0], tail) = os.path.split(sn[0])
        return tail

    assert "test.yml" == _tail(), "nth of resource name is name passed to pkresource"
    assert "package_data" == _tail(), 'n-1th resource is always "package_data"'
    assert "pykern" == _tail(), "n-2th resource is root package of passed in context"
    with pytest.raises(IOError):
        # Should not find somefile, because that's in a different context
        pkresource.filename("somefile", pkresource)
    assert pkresource.filename(
        "somefile", t1.somefile
    ), "Given any object, should fine resource in root package of that object"



# -*- coding: utf-8 -*-
"""Demonstrates a RadiaSoft style module.

This module demonstrates how we code at RadiaSoft.  In general
we follow the `Google Python Style Guide
<http://google.github.io/styleguide/pyguide.html>`_ and the
`Sphinx Napoleon Google Docstrings Example <http://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html>`_.

Some Rules:
    We adhere to PEP8 unless overruled by Google's Style Guide or
    explicit rules such as:

    1. Indent by four (4) spaces and no tabs.
    #. Avoid lines over 80 characters. Not everybody's laptop
       can show two 80+ char pages side by side.
    #. All modules are Python 2 and 3 compatible.
    #. Modules are divided into groups of declarations: public constants,
       private constants, public global variables, private global
       variables, public classes, private classes, public functions,
       and private functions. Within each group, the declarations are
       sorted alphabetically.
    #. Logs, exceptions and assertions contain values, not
       just messages. If a value is in error, include it in the
       message along with the expected value, range, etc.
    #. `:func:pykern.pkdebug.pkdp` to print logging messages,
       and `:func:pykern.pkdebug.pkdc` to print trace messages.
    #. Configuration is specified by `:mod:pykern.pkconfig`.
    #. Use single quotes for strings. Use double quotes for docstrings.
    #. String literals should be concated with ``+`` not by implicit concatenation
    #. TODO(robnagler) is how we represent things to do in a comment

Docstrings begin and end with three double quotes ("). On the line
with the beginning double quotes, you write a one line summary
of the function, class, or module.

:copyright: Copyright (c) 2016 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function

# Imports are sorted alphabetically
# pykern imports are "from" since all modules begin with "pk" so they are unique
from pykern import pkconfig

# Import pkdc and pkdp functions directly. Mostly we import modules and
# qualify uses
from pykern.pkdebug import pkdc, pkdp

# We don't follow PEP8 here. Global variables/constants are separated by one blank
# line.
#: First constant is first alphabetically
ONE = 1

#: Another constant
ZIPPITY = "doodah"

# Private constant is not documented with '#:'
_SSSHHH = 39

#: The module's config is a public variable. It is initialized at the end.
cfg = None

# Something private isn't publicly documented.
_priv_var = 123


# Two blank lines between functions and classes at the global level
class EMA(object):
    """Exponential moving average

    Used as a demonstration of numerical computations. We document
    the __init__ method at the class level, since it is an
    implicit function call.

    Args:
        length (int): iterations

    Attributes:
        average (float): current value of the average
        length (int): time period
    """

    def __init__(self, length):
        self.length = int(length)
        assert length > 0, "{}: length must be greater than 0".format(length)
        self._alpha = 2.0 / (float(length) + 1.0)
        self.average = None

    def compute(self, value):
        """Compute the next value in the average

        Args:
            value (float): next number in the average

        Returns:
            float: current value of the average
        """
        if self.average is None:
            self.average = value
        else:
            self.average += self._alpha * (value - self.average)
        return self.average

    def value(self):
        """Get the average

        Returns:
            float: current value of the average
        """
        assert (
            self.average is not None
        ), "self.average is None and has not been initialized"
        return self.average


def _Privy(object):
    """This is a private class that does nothing"""
    pass


def _cfg_length(anything):
    """Configuration parser for any_length

    Args:
        anything (object): configured value

    Returns:
        int: value between 1 and 999
    """
    anything = int(anything)
    assert 0 < anything <= 999, "{}: any_length must be from 1 to 999"
    return anything


# Finally we assign any length. Note that we include a trailing , at the
# end of every line in a list so that you don't have to remember to
# add the comma when you add another line.
cfg = pkconfig.init(
    any_length=(1, _cfg_length, "A length used by this module"),
)



# -*- coding: utf-8 -*-
"""wrapper for running simulations

:copyright: Copyright (c) 2017 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
from pykern import pkconfig

#: Where we install files with pip
_PYTHON_USER_BASE = "rsbase"

#: how to run python
_PYTHON = ("python", "run.py")

#: how to run bash
_BASH = ("bash", "run.sh")

#: git initialized
_GIT_DIR = ".git"

#: remote host
_GIT_REMOTE = "bitbucket.org"

#: output directory
_OUT_DIR = "out"

#: configuration
cfg = None


def default_command(cmd, *args, **kwargs):
    """Wrapper until figure out args with argh"""
    import sys

    return getattr(sys.modules[__name__], "_cmd_" + cmd)(*args, **kwargs)


def _call(args):
    """Run a command with the proper local python and path environment

    Args:
        args (tuple): what to run (flags and all)
    """
    from pykern import pkio
    import subprocess
    import os

    ub = pkio.py_path(_PYTHON_USER_BASE)
    env = os.environ.copy()
    env["PATH"] = str(ub.join("bin")) + ":" + env["PATH"]
    env["PYTHONUSERBASE"] = str(ub)
    subprocess.check_call(args, env=env)


def _cmd_init(*args):
    """Create git repo locally and on remote"""
    from pykern import pkcli
    import os.path

    # TODO(robnagler) add -public

    if os.path.exists(_GIT_DIR):
        pkcli.command_error("already initialized (.git directory exists)")
    # TODO(robnagler) configure bitbucket locally for each repo
    _init_python_user_base()
    _init_git()


def _cmd_pip(*args):
    """Install a Python package in rsbase

    Args:
        args (tuple): arguments to pass to pip
    """
    args = ["pip", "install", "--user"] + list(args)
    _call(args)
    _git_commit("pip install " + " ".join(args), check_init=True)


def _cmd_run(*args):
    """Execute run.py or run.sh"""
    from pykern import pkcli
    import os.path

    missing = []
    # Prefer _BASH, which may call run.py
    for x in (_BASH, _PYTHON):
        if os.path.exists(x[1]):
            _rsmanifest()
            msg = ": " + " ".join(args) if args else ""
            _git_commit("run" + msg, check_init=True)
            return _call(x)
        missing.append(x[1])
    pkcli.command_error("{}: neither run file exists", missing)


def _git_auth():
    """Get git user.name

    Returns:
        str: configured user name
    """
    from pykern import pkcli
    import netrc

    try:
        b = netrc.netrc().authenticators(_GIT_REMOTE)
        if b:
            return (b[0], b[2])
    except netrc.NetrcParseError:
        pass
    pkcli.command_error('missing login info {}; please "git login"', _GIT_REMOTE)


def _git_commit(msg, check_init=False):
    """Write rsmanifest and commit all files

    Args:
        check_init (bool): make sure git is initialized
    """
    # TODO(robnagler) do every run(?)
    from pykern import pkcli
    import os.path
    import subprocess

    if check_init:
        if not os.path.exists(_GIT_DIR):
            pkcli.command_error('not initialized, please call "init"')
        _git_auth()
    subprocess.check_call(["git", "add", "."])
    subprocess.check_call(["git", "commit", "-m", msg])
    c = ["git", "push"]
    if not check_init:
        c.extend(["-u", "origin", "master"])
    subprocess.check_call(c)


def _git_api_request(method, url, ctx):
    from pykern import pkcli
    import requests

    user, pw = _git_auth()
    ctx["method"] = method
    ctx["user"] = user
    ctx["pass"] = pw
    ctx["host"] = _GIT_REMOTE
    ctx["url"] = ("https://api.{host}/2.0/" + url).format(**ctx)
    x = dict(
        url=ctx["url"],
        method=ctx["method"],
        auth=(user, pw),
    )
    if "json" in ctx:
        x["json"] = ctx["json"]
    r = requests.request(**x)
    # Will return 2xx so best test for now
    if not r.ok:
        pkcli.command_error("{}: post failed: {} {}", ctx["url"], r, r.text)
    return r, ctx


def _init_git():
    """Init git locally and to bitbucket"""
    from pykern import pkcli
    from pykern import pkio
    import datetime
    import re
    import subprocess

    title = pkio.py_path().basename
    v = datetime.datetime.utcnow().strftime("%Y%m%d-%H%M%S")
    name = "sim-{}-{}".format(pkio.py_path().basename, v).lower()
    r, ctx = _git_api_request(
        "post",
        "repositories/{user}/{repo}",
        dict(
            repo=name,
            json=dict(
                scm="git",
                is_private=True,
                fork_policy="no_public_forks",
                name=name,
            ),
        ),
    )
    repo_url = r.json()["links"]["clone"][0]["href"]
    # TODO(robnagler) add README.md if not already there
    subprocess.check_call(["git", "init"])
    subprocess.check_call(["git", "remote", "add", "origin", repo_url])
    subprocess.check_call(["git", "config", "user.name", ctx["user"]])
    if pkio.pkunit_prefix:
        _pkunit_setup(ctx)
    subprocess.check_call(["git", "checkout", "-b", "master"])
    _out_dir()
    _git_commit("init")


def _init_python_user_base():
    """Ensure all python_user_base files are committed"""
    from pykern import pkio

    ub = pkio.py_path(_PYTHON_USER_BASE).ensure_dir()
    ub.join(".gitignore").write("!*\n")


def _out_dir():
    from pykern import pkio

    p = pkio.py_path(_OUT_DIR).ensure_dir()
    p.join(".gitignore").write("*\n!.gitignore\n")


def _pkunit_setup(ctx):
    from pykern import pkio
    import subprocess

    f = pkio.py_path("git-credentials")
    f.write("https://{user}:{pass}@{host}".format(**ctx))
    f.chmod(0o600)
    subprocess.check_call(["git", "config", "credential.helper", "cache"])
    subprocess.check_call(
        ["git", "config", "credential.helper", "store --file " + str(f)]
    )


def _pyenv_version():
    """Determine which pyenv

    Returns:
        str: pyenv version
    """
    import subprocess

    return subprocess.check_output(["pyenv", "version"]).split(" ")[0]


def _rsmanifest():
    from pykern import pkcollections
    from pykern import pkjson
    from pykern.pkcli import rsmanifest
    import cpuinfo
    import datetime
    import os
    import subprocess

    m = rsmanifest.read_all()
    m["sim"] = {
        "run": {
            "datetime": datetime.datetime.utcnow().isoformat(),
            "cpu_info": cpuinfo.get_cpu_info(),
            "pyenv": _pyenv_version(),
            # TODO(robnagler) can't include because of auth/credential
            # values in environment variables
            #'environ': pkcollections.Dict(os.environ),
        },
    }
    pkjson.dump_pretty(m, filename=rsmanifest.BASENAME)



# -*- coding: utf-8 -*-
"""test restartable

:copyright: Copyright (c) 2023 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest


def test_restarable():
    from pykern import pkio
    from pykern import pkunit
    from pykern.pkcli import test

    test._cfg.max_failures = 3
    for test._cfg.restartable in (False, True):
        with pkunit.save_chdir_work() as d:
            pkunit.data_dir().join("tests").copy(d.join("tests"))
            with pkunit.pkexcept(
                "FAILED=1" if test._cfg.restartable else "FAILED=2",
            ):
                test.default_command()



# -*- coding: utf-8 -*-
u"""always passes

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest


def test_2():
    x = 1 / 0



from __future__ import absolute_import, division, print_function

last_cmd = None

last_arg = None

def default_command(arg1):
    global last_cmd, last_arg
    last_cmd = default_command
    last_arg = arg1
    return



# -*- coding: utf-8 -*-
"""Tests error message formatting for config value resolution

:copyright: Copyright (c) 2023 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest


def test_init(pkconfig_setup):
    pkconfig_setup()
    from pykern import pkunit

    with pkunit.pkexcept(r"Error prefix; key=x, value=\(False"):
        from p1.append_error import cfg



# -*- coding: utf-8 -*-
"""MPI support routines

:copyright: Copyright (c) 2017 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function


def checked_call(op):
    """Abort MPI if op raises an exception.

    If ``op`` doesn't handle an exception, then MPI needs
    to abort. This will terminate the MPI process, not just the
    one task.

    If MPI is not running or mpi4py is not installed,
    then passes through the exception.

    Args:
        op (func): function to call
    """
    try:
        op()
    except BaseException as e:
        code = 86
        if isinstance(e, SystemExit) and hasattr(e, "code"):
            if not e.code:
                raise
            code = e.code
        try:
            from mpi4py import MPI

            if MPI.COMM_WORLD and MPI.COMM_WORLD.Get_size() > 1:
                MPI.COMM_WORLD.Abort(code)
        except BaseException as e2:
            pass
        raise



"""test xlsx

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_1():
    from pykern.pkcollections import PKDict
    from pykern.pkdebug import pkdp
    import pykern.pkio
    import pykern.pkrunpy
    import pykern.pkunit
    import sys
    import xml.dom.minidom
    import zipfile

    # If you see a failure, xmllint is helpful:
    #   xmllint --format worksheets/sheet1.xml
    for d in pykern.pkunit.case_dirs():
        with pykern.pkunit.ExceptToFile():
            p = "workbook.xlsx"
            m = pykern.pkrunpy.run_path_as_module("case.py")
            with zipfile.ZipFile(m.PATH, "r") as z:
                z.extractall()


# TODO(robnagler) once we switch to python 3.8+
#                for i in z.infolist():
#                    p = pykern.pkio.py_path(i.filename)
#                    pykern.pkio.mkdir_parent_only(p)
#                    p.write(
#                        xml.dom.minidom.parseString(z.read(i)).toprettyxml(indent='  '),
#                    )
#



# -*- coding: utf-8 -*-
"""pytest for `pykern.pkplatform`

:copyright: Copyright (c) 2015 Bivio Software, Inc.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest


def test_conformance1():
    """Verify the platforms based on other calls"""
    import re
    import sys
    import platform
    from pykern import pkplatform

    if platform.system() == "Linux":
        assert pkplatform.is_linux()
        assert pkplatform.is_unix()
        assert not pkplatform.is_darwin()
        assert not pkplatform.is_windows()
    elif platform.system().startswith("CYGWIN"):
        assert not pkplatform.is_linux()
        assert pkplatform.is_unix()
        assert not pkplatform.is_darwin()
        assert pkplatform.is_windows()
    elif platform.system() == "Windows":
        assert not pkplatform.is_linux()
        assert not pkplatform.is_unix()
        assert not pkplatform.is_darwin()
        assert pkplatform.is_windows()
    elif platform.system() == "Darwin":
        assert not pkplatform.is_linux()
        assert pkplatform.is_unix()
        assert pkplatform.is_darwin()
        assert not pkplatform.is_windows()
    else:
        assert not pkplatform.is_windows()
        assert not pkplatform.is_darwin()
        assert not pkplatform.is_linux()
        # Not sure if it would be unix



"""Wrapper on asyncio and tornado

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern import pkconfig
from pykern.pkdebug import pkdlog, pkdp
import asyncio
import inspect
import tornado.gen
import tornado.httpserver
import tornado.ioloop
import tornado.web

_cfg = None


class Loop:
    def __init__(self):
        _init()
        self._coroutines = []

    def http_server(self, http_cfg):
        """Instantiate a tornado web server

        Under the covers Tornado uses the asyncio event loop so asyncio methods
        can be mixed with Tornado methods.

        To start the server you can call the returned function. Or you can use any of the
        asyncio methods that run the event loop (ex asyncio.run).

        Using asyncio methods (ex asyncio.run()) is prefered over Tornado methods (ex.
        tornado.ioloop.IOLoop.current().start()) to reduce leaking out details of Tornado.

        Args:
            http_cfg (PKDict): object with uri_map and overrides for server defaults
        """

        async def _do():
            # TODO(e-carlin): pull in the one in job_supervisor.py
            p = http_cfg.get("tcp_port", _cfg.server_port)
            i = http_cfg.get("tcp_ip", _cfg.server_ip)
            tornado.httpserver.HTTPServer(
                tornado.web.Application(
                    http_cfg.uri_map,
                    debug=http_cfg.get("debug", _cfg.debug),
                ),
                xheaders=True,
            ).listen(p, i)
            pkdlog("ip={} port={}", p, i)
            await asyncio.Event().wait()

        self.run(_do())

    def run(self, *coros):
        for c in coros:
            if not inspect.iscoroutine(c):
                raise ValueError(f"{c} in coros={coros} is not a coroutine")
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            self._coroutines.extend(coros)
            return
        raise AssertionError("cannot call after event loop has started")

    def start(self):
        async def _do():
            await asyncio.gather(*self._coroutines)

        assert self._coroutines, "nothing to await"
        asyncio.run(_do(), debug=_cfg.debug)


async def sleep(secs):
    await asyncio.sleep(secs)


def _cfg_port(value):
    v = int(value)
    l = 3000
    u = 32767
    if not l <= v <= u:
        pkconfig.raise_error(f"value must be from {l} to {u}")
    return v


def _init():
    global _cfg
    if _cfg:
        return
    _cfg = pkconfig.init(
        debug=(False, bool, "enable debugging for asyncio"),
        server_ip=(
            "0.0.0.0" if pkconfig.in_dev_mode() else "127.0.0.1",
            str,
            "ip to listen on",
        ),
        server_port=("9001", _cfg_port, "port to listen on"),
    )



# -*- coding: utf-8 -*-
"""test pykern.pkcli.sim

:copyright: Copyright (c) 2017 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest
import os


@pytest.mark.skip(reason="need to be able to pass in credentials")
def test_init_and_run(monkeypatch):
    from pykern import pkio
    from pykern import pkunit
    from pykern.pkcli import sim
    from pykern.pkcli import rsmanifest
    import netrc
    import os
    import os.path
    import re
    import subprocess

    # TODO(robnagler) broken: "aux" removed, need another way
    cfg = pkunit.cfg.aux.get("sim_test", None)
    if not cfg:
        # No testing if there's no auth config
        return
    u, p = cfg.split(" ")
    monkeypatch.setattr(netrc, "netrc", _netrc)
    _netrc.result = (u, None, p)
    with pkunit.save_chdir_work(is_pkunit_prefix=True):
        f = "out/log"
        expect_code = pkio.random_string()
        pkio.write_text("run.sh", "echo {}>{}".format(expect_code, f))
        rsmanifest.pkunit_setup()
        sim._cmd_init()
        sim._cmd_run()
        x = (subprocess.check_output(["git", "remote", "-v"]),)
        m = re.search(r"/(sim-sim_work-\d+-\d+)\.git", x[0])
        repo = m.group(1)
        pkunit.pkok(m, "git remote: failed: {}", x)
        pkunit.pkeq(expect_code, pkio.read_text("out/log").rstrip())
        os.remove("run.sh")
        sim._cmd_pip("djson")
        pkio.write_text("run.py", "import djson".format(expect_code, f))
        sim._cmd_run()
        sim._git_api_request(
            "delete",
            "repositories/{user}/{repo}",
            dict(repo=repo),
        )


class _netrc(object):
    def authenticators(self, *args, **kwargs):
        return self.result



# -*- coding: utf-8 -*-
""":mod:`private` package

:copyright: Copyright (c) 2024 James Bond.  All Rights Reserved.
:license: PROPRIETARY AND CONFIDENTIAL. See LICENSE file for details.
"""
import pkg_resources

try:
    # We only have a version once the package is installed.
    __version__ = pkg_resources.get_distribution("private").version
except pkg_resources.DistributionNotFound:
    pass
